# -*- coding: utf-8 -*-
"""
Functions to generate/extract spike trains from analog signals, or to generate
random spike trains.

Extract spike times from time series
***************************************
.. autosummary::
    :toctree: _toctree/spike_train_generation

    spike_extraction
    threshold_detection
    peak_detection


Random spike train processes
****************************
.. autosummary::
    :toctree: _toctree/spike_train_generation
    :template: spike_train_generation_class.rst

    StationaryPoissonProcess
    StationaryGammaProcess
    StationaryLogNormalProcess
    StationaryInverseGaussianProcess
    NonStationaryPoissonProcess
    NonStationaryGammaProcess


Coincident spike times generation
*********************************
.. autosummary::
    :toctree: _toctree/spike_train_generation

    single_interaction_process
    compound_poisson_process

Some functions are based on the NeuroTools stgen module, which was mostly
written by Eilif Muller, or from the NeuroTools signals.analogs module.


References
----------

.. bibliography::
   :keyprefix: generation-

:copyright: Copyright 2014-2023 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division, print_function, unicode_literals

import warnings
from typing import List, Union, Optional

import neo
import numpy as np
import quantities as pq
from scipy import stats
from scipy import integrate
from scipy.optimize import root_scalar
from scipy.special import gammainc, gammaincc

from elephant.spike_train_surrogates import dither_spike_train
from elephant.utils import deprecated_alias

__all__ = [
    "spike_extraction",
    "threshold_detection",
    "peak_detection",
    "StationaryPoissonProcess",
    "StationaryGammaProcess",
    "StationaryLogNormalProcess",
    "StationaryInverseGaussianProcess",
    "NonStationaryPoissonProcess",
    "NonStationaryGammaProcess",
    "homogeneous_poisson_process",
    "inhomogeneous_poisson_process",
    "homogeneous_gamma_process",
    "inhomogeneous_gamma_process",
    "single_interaction_process",
    "compound_poisson_process"
]


@deprecated_alias(extr_interval='interval')
def spike_extraction(signal, threshold=0.0 * pq.mV, sign='above',
                     time_stamps=None, interval=(-2 * pq.ms, 4 * pq.ms)):
    """
    Return the peak times for all events that cross threshold and the
    waveforms. Usually used for extracting spikes from a membrane
    potential to calculate waveform properties.

    Parameters
    ----------
    signal : neo.AnalogSignal
        An analog input signal.
    threshold : pq.Quantity, optional
        Contains a value that must be reached for an event to be detected.
        Default: 0.0 * pq.mV
    sign : {'above', 'below'}, optional
        Determines whether to count threshold crossings that cross above or
        below the threshold.
        Default: 'above'
    time_stamps : pq.Quantity, optional
        If `spike_train` is a `pq.Quantity` array, `time_stamps` provides the
        time stamps around which the waveform is extracted. If it is None, the
        function `peak_detection` is used to calculate the time_stamps
        from signal.
        Default: None
    interval : tuple of pq.Quantity
        Specifies the time interval around the `time_stamps` where the waveform
        is extracted.
        Default: (-2 * pq.ms, 4 * pq.ms)

    Returns
    -------
    result_st : neo.SpikeTrain
        Contains the time_stamps of each of the spikes and the waveforms in
        `result_st.waveforms`.

    See Also
    --------
    elephant.spike_train_generation.peak_detection
    """
    # Get spike time_stamps
    if time_stamps is None:
        time_stamps = peak_detection(signal, threshold, sign=sign)
    elif hasattr(time_stamps, 'times'):
        time_stamps = time_stamps.times
    elif isinstance(time_stamps, pq.Quantity):
        raise TypeError("time_stamps must be None, a pq.Quantity array or" +
                        " expose the.times interface")

    if len(time_stamps) == 0:
        return neo.SpikeTrain(time_stamps, units=signal.times.units,
                              t_start=signal.t_start, t_stop=signal.t_stop,
                              waveforms=np.array([]),
                              sampling_rate=signal.sampling_rate)

    # Unpack the extraction interval from tuple or array
    extr_left, extr_right = interval
    if extr_left > extr_right:
        raise ValueError("interval[0] must be < interval[1]")

    if any(np.diff(time_stamps) < interval[1]):
        warnings.warn("Waveforms overlap.", UserWarning)

    data_left = (extr_left * signal.sampling_rate).simplified.magnitude

    data_right = (extr_right * signal.sampling_rate).simplified.magnitude

    data_stamps = (((time_stamps - signal.t_start) *
                    signal.sampling_rate).simplified).magnitude

    data_stamps = data_stamps.astype(int)

    borders_left = data_stamps + data_left

    borders_right = data_stamps + data_right

    borders = np.dstack((borders_left, borders_right)).flatten()

    waveforms = np.array(
        np.split(np.array(signal), borders.astype(int))[1::2]) * signal.units

    # len(np.shape(waveforms)) == 1 if waveforms do not have the same width.
    # this can occur when extraction interval indexes beyond the signal.
    # Workaround: delete spikes shorter than the maximum length with
    if len(np.shape(waveforms)) == 1:
        max_len = max(len(waveform) for waveform in waveforms)
        to_delete = np.array([idx for idx, x in enumerate(waveforms)
                              if len(x) < max_len])
        waveforms = np.delete(waveforms, to_delete, axis=0)
        warnings.warn("Waveforms " +
                      ("{:d}, " * len(to_delete)).format(*to_delete) +
                      "exceeded signal and had to be deleted. " +
                      "Change 'interval' to keep.")

    waveforms = waveforms[:, np.newaxis, :]

    return neo.SpikeTrain(time_stamps, units=signal.times.units,
                          t_start=signal.t_start, t_stop=signal.t_stop,
                          sampling_rate=signal.sampling_rate,
                          waveforms=waveforms,
                          left_sweep=extr_left)


def threshold_detection(signal, threshold=0.0 * pq.mV, sign='above'):
    """
    Returns the times when the analog signal crosses a threshold.
    Usually used for extracting spike times from a membrane potential.

    Parameters
    ----------
    signal : neo.AnalogSignal
        An analog input signal.
    threshold : pq.Quantity, optional
        Contains a value that must be reached for an event to be detected.
        Default: 0.0 * pq.mV
    sign : {'above', 'below'}, optional
        Determines whether to count threshold crossings that cross above or
        below the threshold.
        Default: 'above'

    Returns
    -------
    result_st : neo.SpikeTrain
        Contains the spike times of each of the events (spikes) extracted from
        the signal.
    """

    if not isinstance(threshold, pq.Quantity):
        raise ValueError('threshold must be a pq.Quantity')

    if sign not in ('above', 'below'):
        raise ValueError("sign should be 'above' or 'below'")

    if sign == 'above':
        cutout = np.where(signal > threshold)[0]
    else:
        # sign == 'below'
        cutout = np.where(signal < threshold)[0]

    if len(cutout) == 0:
        events_base = np.zeros(0)
    else:
        take = np.where(np.diff(cutout) > 1)[0] + 1
        take = np.append(0, take)

        time = signal.times
        events = time[cutout][take]

        events_base = events.magnitude
        if events_base is None:
            # This occurs in some Python 3 builds due to some
            # bug in quantities.
            events_base = np.array(
                [event.magnitude for event in events])  # Workaround

    result_st = neo.SpikeTrain(events_base, units=signal.times.units,
                               t_start=signal.t_start, t_stop=signal.t_stop)
    return result_st


@deprecated_alias(format='as_array')
def peak_detection(signal, threshold=0.0 * pq.mV, sign='above',
                   as_array=False):
    """
    Return the peak times for all events that cross threshold.
    Usually used for extracting spike times from a membrane potential.
    Similar to spike_train_generation.threshold_detection.

    Parameters
    ----------
    signal : neo.AnalogSignal
        An analog input signal.
    threshold : pq.Quantity, optional
        Contains a value that must be reached for an event to be detected.
        Default: 0.*pq.mV
    sign : {'above', 'below'}, optional
        Determines whether to count threshold crossings that cross above or
        below the threshold.
        Default: 'above'
    as_array : bool, optional
        If True, a NumPy array of the resulting peak times is returned instead
        of a (default) `neo.SpikeTrain` object.
        Default: False
    format : {None, 'raw'}, optional

        .. deprecated:: 0.8.0

        Whether to return as SpikeTrain (None) or as a plain array of times
        ('raw').
        Deprecated. Use `as_array=False` for None format and `as_array=True`
        otherwise.
        Default: None

    Returns
    -------
    result_st : neo.SpikeTrain
        Contains the spike times of each of the events (spikes) extracted from
        the signal.
    """
    if not isinstance(threshold, pq.Quantity):
        raise ValueError("threshold must be a pq.Quantity")

    if sign not in ('above', 'below'):
        raise ValueError("sign should be 'above' or 'below'")

    if as_array in (None, 'raw'):
        warnings.warn("'format' is deprecated; use as_array=True",
                      DeprecationWarning)
        as_array = bool(as_array)

    if sign == 'above':
        cutout = np.where(signal > threshold)[0]
        peak_func = np.argmax
    else:
        # sign == 'below'
        cutout = np.where(signal < threshold)[0]
        peak_func = np.argmin

    if len(cutout) == 0:
        events_base = np.zeros(0)
    else:
        # Select thr crossings lasting at least 2 dtps, np.diff(cutout) > 2
        # This avoids empty slices
        border_start = np.where(np.diff(cutout) > 1)[0]
        border_end = border_start + 1
        borders = sorted(np.r_[0, border_start, border_end, len(cutout) - 1])
        true_borders = cutout[borders]
        right_borders = true_borders[1::2] + 1
        true_borders = np.sort(np.append(true_borders[0::2], right_borders))

        # Workaround for bug that occurs when signal goes below thr for 1 dtp,
        # Workaround eliminates empty slices from np. split
        backward_mask = np.absolute(np.ediff1d(true_borders, to_begin=1)) > 0
        forward_mask = np.absolute(np.ediff1d(true_borders[::-1],
                                              to_begin=1)[::-1]) > 0
        true_borders = true_borders[backward_mask * forward_mask]
        split_signal = np.split(np.array(signal), true_borders)[1::2]

        maxima_idc_split = np.array([peak_func(x) for x in split_signal])

        max_idc = maxima_idc_split + true_borders[0::2]

        events = signal.times[max_idc]
        events_base = events.magnitude

        if events_base is None:
            # This occurs in some Python 3 builds due to some
            # bug in quantities.
            events_base = np.array(
                [event.magnitude for event in events])  # Workaround

    result_st = neo.SpikeTrain(events_base, units=signal.times.units,
                               t_start=signal.t_start,
                               t_stop=signal.t_stop)
    if as_array:
        result_st = result_st.magnitude

    return result_st


class AbstractPointProcess:
    """
    Abstract point process to subclass from.

    Parameters
    ----------
    t_start : pq.Quantity, optional
        The start of the spike train.
        Default: 0.*pq.s
    t_stop : pq.Quantity, optional
        The end of the spike train.
        Default: 1.*pq.s
    """
    def __init__(
            self,
            t_stop: pq.Quantity = 1.*pq.s,
            t_start: pq.Quantity = 0.*pq.s
    ):
        if not (isinstance(t_start, pq.Quantity) and
                isinstance(t_stop, pq.Quantity)):
            raise ValueError("t_start and t_stop must be of type pq.Quantity")
        if t_stop <= t_start:
            raise ValueError('t_start must be smaller than t_stop.')

        self.units = t_stop.units
        self._t_stop = t_stop.item()
        self._t_start = t_start.rescale(self.units).item()

    @property
    def t_start(self):
        """
        t_start quantity; there are no spike times below this value.
        """
        return self._t_start * self.units

    @property
    def t_stop(self):
        """
        t_stop quantity; there are no spike times above this value.
        """
        return self._t_stop * self.units

    def _generate_spiketrain_as_array(self) -> np.ndarray:
        raise NotImplementedError

    def generate_spiketrain(
            self, as_array: bool = False
    ) -> Union[neo.SpikeTrain, np.ndarray]:
        """
        Generates a single spike train.

        Parameters
        ----------
        as_array : bool, optional
            If True, a NumPy array of sorted spikes is returned,
            rather than a `neo.SpikeTrain` object.
            Default: False

        Returns
        -------
        spiketrain : neo.SpikeTrain or np.ndarray
            The generated spike train in the specified format.
        """
        spikes = self._generate_spiketrain_as_array()
        if as_array:
            return spikes
        # else:
        return neo.SpikeTrain(
            spikes,
            t_start=self.t_start, t_stop=self.t_stop, units=self.units)

    def generate_n_spiketrains(
            self,
            n_spiketrains: int,
            as_array: bool = False
    ) -> Union[List[neo.SpikeTrain], List[np.ndarray]]:
        """
        Generates a list of spike trains.

        Parameters
        ----------
        n_spiketrains : int
            The number of spike trains to generate.
        as_array : bool, optional
            If True, a NumPy array of sorted spikes is returned,
            rather than a `neo.SpikeTrain` object.
            Default: False

        Returns
        -------
        list_of_spiketrain : list of neo.SpikeTrain or list of np.ndarray
            A list generated spike trains in the specified format.
        """
        return [self.generate_spiketrain(as_array=as_array)
                for _ in range(n_spiketrains)]


class RenewalProcess(AbstractPointProcess):
    """
    Abstract renewal process to subclass from.

    Parameters
    ----------
    rate : pq.Quantity
        The constant firing rate.
    t_start : pq.Quantity, optional
        The start of the spike train.
        Default: 0.*pq.s
    t_stop : pq.Quantity, optional
        The end of the spike train.
        Default: 1.*pq.s
    equilibrium : bool, optional
        Generate an equilibrium or an ordinary renewal process.
        Default: True
    """
    isi_generator: stats.rv_continuous

    def __init__(
            self,
            rate: pq.Quantity,
            t_start: pq.Quantity = 0.*pq.s,
            t_stop: pq.Quantity = 1.*pq.s,
            equilibrium: bool = True
    ):
        super().__init__(t_start=t_start, t_stop=t_stop)
        if not isinstance(rate, pq.Quantity):
            raise ValueError("rate must be of type pq.Quantity")
        self.rate = rate.rescale(1./self.units).item()

        self.equilibrium = equilibrium

        self.n_expected_spikes = int(np.ceil(
            ((self._t_stop - self._t_start) * self.rate)))

        if self.n_expected_spikes < 0:
            raise ValueError(
                f"Expected no. of spikes: {self.n_expected_spikes} < 0. "
                f"The firing rate ({self.rate/self.units}) "
                f"cannot be negative.")

    def _cdf_first_spike_equilibrium(self, time):
        """
        Integral over the p.d.f. of the first spike which is:
        p(t) = rate * survival-function(t) * Heaviside(t).
        See Bouss (2020).

        The parameter time is a magnitude of a time value given in seconds.
        """
        return self.rate * integrate.quad(self.isi_generator.sf, 0., time)[0]

    def _get_first_spike_equilibrium(self):
        """
        Return a numerically drawn sample of the p.d.f of the first spike.

        By solving:
        x = integral(c.d.f(t) from 0 to t),
        where x is drawn from a uniform distribution.
        """
        random_uniform = np.random.random()
        equation_solver = root_scalar

        def function_to_solve(time):
            """
            # integral(c.d.f(t) from 0 to t) - random-number-x)
            """
            return self._cdf_first_spike_equilibrium(time) - random_uniform

        def derivative_of_function_to_solve(time):
            """
            derivative of the c.d.f, which is rate times
            the survival function
            """
            return self.rate * self.isi_generator.sf(time)

        # Initial guess is solution for Poisson process
        initial_guess = -np.log(1.-random_uniform)/self.rate
        duration = self._t_stop-self._t_start
        limits_for_first_spike = (0., duration)

        # test if solution for first spike is inside the boundaries. If not
        # return t_stop of the spike train.
        if self._cdf_first_spike_equilibrium(duration) <= random_uniform:
            return self._t_stop

        non_shifted_position_of_first_spike = equation_solver(
                function_to_solve,
                x0=initial_guess,
                bracket=limits_for_first_spike,
                fprime=derivative_of_function_to_solve
            ).root

        return non_shifted_position_of_first_spike + self._t_start

    def _generate_spiketrain_as_array(self) -> np.ndarray:
        if self.n_expected_spikes == 0:
            return np.array([])

        if self.equilibrium:  # equilibrium renewal process
            # First spike of equilibrium renewal process drawn according to
            # Bouss (2020), Master's Thesis
            first_spike = self._get_first_spike_equilibrium()
        else:  # ordinary renewal process
            first_spike = self.isi_generator.rvs() + self._t_start

        spikes = np.array([first_spike])

        # 3 STDs corresponds to 99.7%
        n_spikes_three_stds = int(np.ceil(
            self.n_expected_spikes + 3 * np.sqrt(self.n_expected_spikes)))

        # Continue until whole time range is covered
        while spikes[-1] < self._t_stop:
            isi = self.isi_generator.rvs(size=n_spikes_three_stds)

            t_last_spike = spikes[-1]
            spikes = np.r_[spikes, t_last_spike + np.cumsum(isi)]

        index_last_spike = spikes.searchsorted(self._t_stop)
        spikes = spikes[:index_last_spike]

        return spikes

    @property
    def expected_cv(self):
        """
        The expected coefficient of variation given the ISI distribution.
        """
        return self.isi_generator.std()/self.isi_generator.mean()


class StationaryPoissonProcess(RenewalProcess):
    """
    Generates spike trains whose spikes are realizations of a stationary
    Poisson process with the given rate, starting at time `t_start` and
    stopping at time `t_stop` :cite:`generation-Deger12_443`. Optionally,
    an absolute refractory period / dead time can be specified.

    Parameters
    ----------
    rate : pq.Quantity
        The constant firing rate.
    t_start : pq.Quantity, optional
        The start of the spike train.
        Default: 0.*pq.s
    t_stop : pq.Quantity, optional
        The end of the spike train.
        Default: 1.*pq.s
    refractory_period : pq.Quantity, optional
        The time period after one spike in which no other spike is emitted.
        This can be called an absolute refractory period or a dead time as
        used in :cite:`generation-Deger12_443`.
        Default : None
    equilibrium : bool, optional
        Generate an equilibrium or an ordinary renewal process.
        Default: True

    Raises
    ------
    ValueError
        If one of `rate`, `t_start` and `t_stop` is not of type `pq.Quantity`.

        If `refractory_period` is not of type `pq.Quantity` nor None.

        If the period between two successive spikes (`1 / rate`) is smaller
        or equal than the `refractory_period`.

    Examples
    --------
    >>> import quantities as pq
    >>> spiketrain = StationaryPoissonProcess(rate=50.*pq.Hz, t_start=0*pq.ms,
    ...     t_stop=1000*pq.ms).generate_spiketrain()
    >>> spiketrain_array = StationaryPoissonProcess(
    ...     rate=20*pq.Hz, t_start=5000*pq.ms, t_stop=10000*pq.ms
    ...     ).generate_spiketrain(as_array=True)
    >>> spiketrain = StationaryPoissonProcess(
    ...     rate=50*pq.Hz,
    ...     t_start=0*pq.ms, t_stop=1000*pq.ms,
    ...     refractory_period = 3*pq.ms).generate_spiketrain()
    """
    def __init__(
            self,
            rate: pq.Quantity,
            t_stop: pq.Quantity = 1.*pq.s,
            t_start: pq.Quantity = 0.*pq.s,
            refractory_period: Optional[pq.Quantity] = None,
            equilibrium: bool = True
    ):
        super().__init__(
            rate=rate, t_start=t_start, t_stop=t_stop, equilibrium=equilibrium)

        if refractory_period is not None:
            if not isinstance(refractory_period, pq.Quantity):
                raise ValueError(
                    "refractory_period must be of type pq.Quantity")
            self.refractory_period = refractory_period.rescale(
                self.units).item()

            if self.rate * self.refractory_period >= 1.:
                raise ValueError(
                    "Period between two successive spikes must be larger "
                    "than the refractory period. Decrease either the "
                    "firing rate or the refractory period.")
        else:
            self.refractory_period = refractory_period

        if self.n_expected_spikes > 0 and refractory_period is None:
            self.isi_generator = stats.expon(scale=1./self.rate)

        elif self.n_expected_spikes > 0 and refractory_period is not None:
            self.effective_rate = self.rate / \
                                  (1. - self.rate * self.refractory_period)
            self.isi_generator = stats.expon(
                scale=1. / self.effective_rate, loc=self.refractory_period)

    def _get_first_spike_equilibrium(self):
        if self.refractory_period is None:
            return self.isi_generator.rvs() + self._t_start

        # the case with dead time
        random_uniform = np.random.random()
        if random_uniform <= self.rate * self.refractory_period:
            return random_uniform / self.rate + self._t_start
        # random_uniform > self.rate * self.refractory_period
        return (np.log(1. - self.rate * self.refractory_period)
                - np.log(1. - random_uniform)
                ) / self.effective_rate + self.refractory_period

    @property
    def expected_cv(self):
        """
        The expected coefficient of variation given the ISI distribution.
        """
        if self.refractory_period is None:
            return 1.

        # the case with dead time
        return 1. - self.rate * self.refractory_period


class StationaryGammaProcess(RenewalProcess):
    """
    Generates spike trains whose spikes are realizations of a stationary
    Gamma process with the given rate and `shape_factor`
    starting at time `t_start` and stopping at time `t_stop`.

    Parameters
    ----------
    rate : pq.Quantity
        The constant firing rate.
    shape_factor : float
        The shape parameter of the gamma distribution.
    t_start : pq.Quantity, optional
        The start of the spike train.
        Default: 0.*pq.s
    t_stop : pq.Quantity, optional
        The end of the spike train.
        Default: 1.*pq.s
    equilibrium : bool, optional
        Generate an equilibrium or an ordinary renewal process.
        Default: True

    Raises
    ------
    ValueError
        If one of `rate`, `t_start` and `t_stop` is not of type `pq.Quantity`.

    Examples
    --------
    >>> import quantities as pq
    >>> spiketrain = StationaryGammaProcess(
    ...     rate=50*pq.Hz, shape_factor=2.0, t_start=0*pq.ms,
    ...     t_stop=1000*pq.ms).generate_spiketrain()
    >>> spiketrain_array = StationaryGammaProcess(
    ...     rate=20*pq.Hz, shape_factor=5.0, t_start=5000*pq.ms,
    ...     t_stop=10000*pq.ms).generate_spiketrain(as_array=True)
    """
    def __init__(
            self,
            rate: pq.Quantity,
            shape_factor: float,
            t_start: pq.Quantity = 0.*pq.s,
            t_stop: pq.Quantity = 1.*pq.s,
            equilibrium: bool = True
    ):
        super().__init__(
            rate=rate, t_start=t_start, t_stop=t_stop, equilibrium=equilibrium)
        if self.n_expected_spikes > 0:
            self.shape_factor = shape_factor
            self.isi_generator = stats.gamma(
                a=shape_factor, scale=1./(shape_factor * self.rate))

    def _cdf_first_spike_equilibrium(self, time):
        """
        The parameter time is a magnitude of a time value given in seconds.
        """
        if time < 0.:
            return 0.
        return self.rate * time * \
            gammaincc(self.shape_factor,
                      self.shape_factor*self.rate*time)\
            + gammainc(self.shape_factor+1.,
                       self.shape_factor*self.rate*time)

    @property
    def expected_cv(self):
        """
        The expected coefficient of variation given the ISI distribution.
        """
        return 1./np.sqrt(self.shape_factor)


class StationaryLogNormalProcess(RenewalProcess):
    """
    Generates spike trains whose spikes are realizations of a stationary
    LogNormal process with the given rate and `sigma`
    starting at time `t_start` and stopping at time `t_stop`.

    Parameters
    ----------
    rate : pq.Quantity
        The constant firing rate.
    sigma : float
        The sigma/ s parameter of the Log-Normal distribution.
    t_start : pq.Quantity, optional
        The start of the spike train.
        Default: 0.*pq.s
    t_stop : pq.Quantity, optional
        The end of the spike train.
        Default: 1.*pq.s
    equilibrium : bool, optional
        Generate an equilibrium or an ordinary renewal process.
        Default: True

    Raises
    ------
    ValueError
        If one of `rate`, `t_start` and `t_stop` is not of type `pq.Quantity`.

    Examples
    --------
    >>> import quantities as pq
    >>> spiketrain = StationaryLogNormalProcess(
    ...     rate=50*pq.Hz, sigma=2.0, t_start=0*pq.ms,
    ...     t_stop=1000*pq.ms).generate_spiketrain()
    >>> spiketrain_array = StationaryLogNormalProcess(
    ...     rate=20*pq.Hz, sigma=5.0, t_start=5000*pq.ms,
    ...     t_stop=10000*pq.ms).generate_spiketrain(as_array=True)
    """
    def __init__(
            self,
            rate: pq.Quantity,
            sigma: float,
            t_start: pq.Quantity = 0.*pq.s,
            t_stop: pq.Quantity = 1.*pq.s,
            equilibrium: bool = True
    ):
        super().__init__(
            rate=rate, t_start=t_start, t_stop=t_stop, equilibrium=equilibrium)
        self.sigma = sigma
        if self.n_expected_spikes > 0:
            self.isi_generator = stats.lognorm(
                s=self.sigma, scale=np.exp(self.mu))

    @property
    def mu(self):
        """
        The parameter mu of the log-normal distribution.
        """
        return -np.log(self.rate) - self.sigma**2/2

    @property
    def expected_cv(self):
        """
        The expected coefficient of variation given the ISI distribution.
        """
        return np.sqrt(np.exp(self.sigma**2) - 1)


class StationaryInverseGaussianProcess(RenewalProcess):
    """
    Generates spike trains whose spikes are realizations of a stationary
    Gamma process with the given rate and `cv`
    starting at time `t_start` and stopping at time `t_stop`.

    Raises
    ------
    ValueError
        If one of `rate`, `t_start` and `t_stop` is not of type `pq.Quantity`.

    Parameters
    ----------
    rate : pq.Quantity
        The constant firing rate.
    cv : float
        The expected coefficient of variation.
    t_start : pq.Quantity, optional
        The start of the spike train.
        Default: 0.*pq.s
    t_stop : pq.Quantity, optional
        The end of the spike train.
        Default: 1.*pq.s
    equilibrium : bool, optional
        Generate an equilibrium or an ordinary renewal process.
        Default: True

    Examples
    --------
    >>> import quantities as pq
    >>> spiketrain = StationaryInverseGaussianProcess(
    ...     rate=50*pq.Hz, cv=2.0, t_start=0*pq.ms,
    ...     t_stop=1000*pq.ms).generate_spiketrain()
    >>> spiketrain_array = StationaryInverseGaussianProcess(
    ...     rate=20*pq.Hz, cv=5.0, t_start=5000*pq.ms,
    ...     t_stop=10000*pq.ms).generate_spiketrain(as_array=True)
    """
    def __init__(
            self,
            rate: pq.Quantity,
            cv: float,
            t_start: pq.Quantity = 0.*pq.s,
            t_stop: pq.Quantity = 1.*pq.s,
            equilibrium: bool = True
    ):
        super().__init__(
            rate=rate, t_start=t_start, t_stop=t_stop, equilibrium=equilibrium)
        self._cv = cv
        if self.n_expected_spikes > 0:
            self.isi_generator = stats.invgauss(
                mu=cv**2, scale=1./(self.rate*cv**2))

    @property
    def expected_cv(self):
        """
        The expected coefficient of variation given the ISI distribution.
        """
        return self._cv


class RateModulatedProcess(AbstractPointProcess):
    """
    Abstract rate-modulated process to subclass from.

    Parameters
    ----------
    rate_signal : neo.AnalogSignal
        A `neo.AnalogSignal` representing the rate profile evolving over
        time. Its values have all to be `>=0`. The generated spike trains
        will have `t_start = rate.t_start` and `t_stop = rate.t_stop`

    Raises
    ------
    ValueError
        If `rate_signal` is not a neo AnalogSignal
        If `rate_signal` contains a negative value.
        If `rate_signal` is empty.
    """
    process_operational_time: RenewalProcess

    def __init__(self, rate_signal: neo.AnalogSignal):

        if not isinstance(rate_signal, neo.AnalogSignal):
            raise ValueError(
                f'rate_signal should be of type neo.AnalogSignal.'
                f' Currently it is of type: {type(rate_signal)}')
        if len(rate_signal) == 0:
            raise ValueError('rate_signal can not be empty.')
        if any(rate_signal < 0):
            raise ValueError(
                'All elements of rate_signal should be positive.')

        super().__init__(
            t_start=rate_signal.t_start, t_stop=rate_signal.t_stop)

        self.rate_signal = rate_signal

        self.mean_rate = np.mean(rate_signal.rescale(1./self.units).magnitude)

        if self.mean_rate == 0.:
            # if the firing rate is zero, the init functions stops here, since
            # the other parameters are then not needed.
            return None

        self.sampling_period = \
            self.rate_signal.sampling_period.rescale(self.units).magnitude
        # Operational time corresponds to the integral of the firing rate
        # over time, here normalized by the average firing rate
        operational_time = np.cumsum(
            rate_signal.rescale(1./self.units).magnitude)
        operational_time *= (self.sampling_period / self.mean_rate)
        operational_time = np.hstack((0., operational_time))
        self.operational_time = operational_time + self._t_start

        # The time points at which the firing rates are given
        self.real_time = np.hstack(
            (rate_signal.times.rescale(self.units).magnitude,
             self._t_stop))

    def _generate_spiketrain_as_array(self) -> np.ndarray:
        spiketrain_operational_time = \
            self.process_operational_time._generate_spiketrain_as_array()
        if len(spiketrain_operational_time) == 0:
            return spiketrain_operational_time
        # indices where between which points in operational time the spikes lie
        indices = np.searchsorted(self.operational_time,
                                  spiketrain_operational_time)

        # In real time the spikes are first aligned
        # to the left border of the bin.
        # Note that indices are greater than 0 because 'operational_time' was
        # padded with zeros.
        spiketrain = self.real_time[indices - 1]
        # the relative position of the spikes in the operational time bins
        positions_in_bins = \
            (spiketrain_operational_time
             - self.operational_time[indices - 1]) / \
            (self.operational_time[indices]
             - self.operational_time[indices - 1])

        # add the positions in the bin times the sampling period in real time
        spiketrain += self.sampling_period * positions_in_bins
        return spiketrain


class NonStationaryPoissonProcess(RateModulatedProcess):
    """
    Generates spike trains whose spikes are realizations of a non-stationary
    Poisson process with the given `rate-signal`. Optionally, you can specify a
    dead time.

    Parameters
    ----------
    rate_signal : neo.AnalogSignal
        A `neo.AnalogSignal` representing the rate profile evolving over
        time.Its values have all to be `>=0`. The generated spike trains
        will have `t_start = rate.t_start` and `t_stop = rate.t_stop`
    refractory_period : pq.Quantity, optional
        The time period after one spike in which no other spike is emitted.
        This can be called an absolute refractory period or a dead time.
        Default : None

    Raises
    ------
    ValueError
        If `rate_signal` is not a neo AnalogSignal
        If `rate_signal` contains a negative value.
        If `rate_signal` is empty.
        If `refractory_period` is not of type `pq.Quantity` nor None.
    """
    def __init__(self, rate_signal: neo.AnalogSignal,
                 refractory_period: Optional[pq.Quantity] = None):

        if refractory_period is not None:
            if not isinstance(refractory_period, pq.Quantity):
                raise ValueError(
                    "refractory_period must be of type pq.Quantity")
            rate_signal = \
                rate_signal / (1. - rate_signal.simplified.magnitude
                               * refractory_period.simplified.item())

        super().__init__(rate_signal=rate_signal)
        self.process_operational_time = StationaryPoissonProcess(
            rate=self.mean_rate * 1./self.units,
            t_start=self.t_start,
            t_stop=self.t_stop)

        self.refractory_period = refractory_period
        if self.refractory_period is not None:
            self.refractory_period = self.refractory_period.rescale(
                self.units).item()

    def _generate_spiketrain_as_array(self) -> np.ndarray:
        if self.refractory_period is None:
            return super()._generate_spiketrain_as_array()

        spiketrain = super()._generate_spiketrain_as_array()
        thinned_spiketrain = []

        previous_spike = self._t_start - self.refractory_period

        for spike in spiketrain:
            if spike > previous_spike + self.refractory_period:
                thinned_spiketrain.append(spike)
                previous_spike = spike
        return np.array(thinned_spiketrain)


class NonStationaryGammaProcess(RateModulatedProcess):
    """
    Generates spike trains whose spikes are realizations of a non-stationary
    Gamma process with the given `rate-signal`.

    Parameters
    ----------
    rate_signal : neo.AnalogSignal
        A `neo.AnalogSignal` representing the rate profile evolving over
        time.Its values have all to be `>=0`. The generated spike trains
        will have `t_start = rate.t_start` and `t_stop = rate.t_stop`
    shape_factor : float
        The shape parameter of the gamma distribution.

    Raises
    ------
    ValueError
        If `rate_signal` is not a neo AnalogSignal
        If `rate_signal` contains a negative value.
        If `rate_signal` is empty.
    """
    def __init__(self, rate_signal: neo.AnalogSignal, shape_factor: float):
        super().__init__(rate_signal=rate_signal)
        self.process_operational_time = StationaryGammaProcess(
            rate=self.mean_rate * 1./self.units,
            shape_factor=shape_factor,
            t_start=self.t_start,
            t_stop=self.t_stop)


def homogeneous_poisson_process(rate, t_start=0.0 * pq.ms,
                                t_stop=1000.0 * pq.ms, as_array=False,
                                refractory_period=None):
    """
    Returns a spike train whose spikes are a realization of a Poisson process
    with the given rate, starting at time `t_start` and stopping time `t_stop`.

    All numerical values should be given as Quantities, e.g. `100*pq.Hz`.

    Parameters
    ----------
    rate : pq.Quantity
        The rate of the discharge.
    t_start : pq.Quantity, optional
        The beginning of the spike train.
        Default: 0 * pq.ms
    t_stop : pq.Quantity, optional
        The end of the spike train.
        Default: 1000 * pq.ms
    as_array : bool, optional
        If True, a NumPy array of sorted spikes is returned,
        rather than a `neo.SpikeTrain` object.
        Default: False
    refractory_period : pq.Quantity or None, optional
        `pq.Quantity` scalar with dimension time. The time period after one
        spike no other spike is emitted.
        Default: None

    Returns
    -------
    spiketrain : neo.SpikeTrain or np.ndarray
        Homogeneous Poisson process realization, stored in `neo.SpikeTrain`
        if `as_array` is False (default) and `np.ndarray` otherwise.

    Raises
    ------
    ValueError
        If one of `rate`, `t_start` and `t_stop` is not of type `pq.Quantity`.

        If `refractory_period` is not None or not of type `pq.Quantity`.

        If `refractory_period` is not None and the period between two
        successive spikes (`1 / rate`) is smaller than the `refractory_period`.

    Examples
    --------
    >>> import quantities as pq
    >>> spikes = homogeneous_poisson_process(50*pq.Hz, t_start=0*pq.ms,
    ...     t_stop=1000*pq.ms)
    >>> spikes = homogeneous_poisson_process(
    ...     20*pq.Hz, t_start=5000*pq.ms, t_stop=10000*pq.ms, as_array=True)
    >>> spikes = homogeneous_poisson_process(50*pq.Hz, t_start=0*pq.ms,
    ...     t_stop=1000*pq.ms, refractory_period = 3*pq.ms)

    """
    warnings.warn(
        "'homogeneous_poisson_process' is deprecated;"
        " use 'StationaryPoissonProcess'.", DeprecationWarning)
    process = StationaryPoissonProcess(
        rate=rate, t_start=t_start, t_stop=t_stop,
        refractory_period=refractory_period, equilibrium=False)
    return process.generate_spiketrain(as_array=as_array)


def inhomogeneous_poisson_process(rate, as_array=False,
                                  refractory_period=None):
    """
    Returns a spike train whose spikes are a realization of an inhomogeneous
    Poisson process with the given rate profile.

    Parameters
    ----------
    rate : neo.AnalogSignal
        A `neo.AnalogSignal` representing the rate profile evolving over time.
        Its values have all to be `>=0`. The output spiketrain will have
        `t_start = rate.t_start` and `t_stop = rate.t_stop`
    as_array : bool, optional
        If True, a NumPy array of sorted spikes is returned,
        rather than a SpikeTrain object.
        Default: False
    refractory_period : pq.Quantity or None, optional
        `pq.Quantity` scalar with dimension time. The time period after one
        spike no other spike is emitted.
        Default: None

    Returns
    -------
    spiketrain : neo.SpikeTrain or np.ndarray
        Inhomogeneous Poisson process realization, of type `neo.SpikeTrain`
        if `as_array` is False (default) and `np.ndarray` otherwise.

    Raises
    ------
    ValueError
        If `rate` contains a negative value.

        If `refractory_period` is not None or not of type `pq.Quantity`.

        If `refractory_period` is not None and the period between two
        successive spikes (`1 / rate`) is smaller than the `refractory_period`.

    """
    warnings.warn(
        "'inhomogeneous_poisson_process' is deprecated;"
        " use 'NonStationaryPoissonProcess'.",
        DeprecationWarning)
    process = NonStationaryPoissonProcess(
        rate_signal=rate,
        refractory_period=refractory_period)
    return process.generate_spiketrain(
            as_array=as_array)


def homogeneous_gamma_process(a, b, t_start=0.0 * pq.ms, t_stop=1000.0 * pq.ms,
                              as_array=False):
    """
    Returns a spike train whose spikes are a realization of a gamma process
    with the given parameters, starting at time `t_start` and stopping time
    `t_stop` (average rate will be `b/a`). All numerical values should be
    given as Quantities, e.g. `100*pq.Hz`.

    Parameters
    ----------
    a : int or float
        The shape parameter of the gamma distribution.
    b : pq.Quantity
        The rate parameter of the gamma distribution.
    t_start : pq.Quantity, optional
        The beginning of the spike train.
        Default: 0 * pq.ms
    t_stop : pq.Quantity, optional
        The end of the spike train.
        Default: 1000 * pq.ms
    as_array : bool, optional
        If True, a NumPy array of sorted spikes is returned, rather than a
        `neo.SpikeTrain` object.
        Default: False

    Returns
    -------
    spiketrain : neo.SpikeTrain or np.ndarray
        Homogeneous Gamma process realization, stored in `neo.SpikeTrain`
        if `as_array` is False (default) and `np.ndarray` otherwise.

    Raises
    ------
    ValueError
        If `t_start` and `t_stop` are not of type `pq.Quantity`.

    Examples
    --------
    >>> import quantities as pq
    >>> spikes = homogeneous_gamma_process(2.0, 50*pq.Hz, 0*pq.ms,
    ...                                       1000*pq.ms)
    >>> spikes = homogeneous_gamma_process(
    ...        5.0, 20*pq.Hz, 5000*pq.ms, 10000*pq.ms, as_array=True)

    """
    warnings.warn(
        "'homogeneous_gamma_process' is deprecated;"
        " use 'StationaryGammaProcess'.",
        DeprecationWarning)
    process = StationaryGammaProcess(
        rate=b / a, shape_factor=a, t_stop=t_stop, t_start=t_start,
        equilibrium=False)
    return process.generate_spiketrain(as_array=as_array)


def inhomogeneous_gamma_process(rate, shape_factor, as_array=False):
    """
    Returns a spike train whose spikes are a realization of an inhomogeneous
    Gamma process with the given rate profile and the given shape factor
    :cite:`generation-Nawrot2008_374`.

    Parameters
    ----------
    rate : neo.AnalogSignal
        A `neo.AnalogSignal` representing the rate profile evolving over time.
        Its values have all to be `>=0`. The output spiketrain will have
        `t_start = rate.t_start` and `t_stop = rate.t_stop`
    shape_factor : float
        The shape factor of the Gamma process
    as_array : bool, optional
        If True, a NumPy array of sorted spikes is returned,
        rather than a SpikeTrain object.
        Default: False

    Returns
    -------
    spiketrain : neo.SpikeTrain or np.ndarray
        Inhomogeneous Poisson process realization, of type `neo.SpikeTrain`
        if `as_array` is False (default) and `np.ndarray` otherwise.

    Raises
    ------
    ValueError
        If `rate` is not a neo AnalogSignal
        If `rate` contains a negative value.

    """
    warnings.warn(
        "'inhomogeneous_gamma_process' is deprecated;"
        " use 'nonStationaryGammaProcess'.",
        DeprecationWarning)
    process = NonStationaryGammaProcess(
        rate_signal=rate, shape_factor=shape_factor)
    return process.generate_spiketrain(as_array=as_array)


@deprecated_alias(n='n_spiketrains')
def _n_poisson(rate, t_stop, t_start=0.0 * pq.ms, n_spiketrains=1):
    """
    Generates one or more independent Poisson spike trains.

    Parameters
    ----------
    rate : pq.Quantity scalar or pq.Quantity array
        Expected firing rate (frequency) of each output SpikeTrain.
        Can be one of:
        *  a single pq.Quantity value: expected firing rate of each output
           SpikeTrain
        *  a pq.Quantity array: rate[i] is the expected firing rate of the i-th
           output SpikeTrain
    t_stop : pq.Quantity
        Single common stop time of each output SpikeTrain. Must be > t_start.
    t_start : pq.Quantity, optional
        Single common start time of each output SpikeTrain. Must be < t_stop.
        Default: 0 * pq.ms
    n_spiketrains : int, optional
        If rate is a single pq.Quantity value, n specifies the number of
        SpikeTrains to be generated. If rate is an array, n is ignored and the
        number of SpikeTrains is equal to len(rate).
        Default: 1


    Returns
    -------
    list of neo.SpikeTrain
        Each SpikeTrain contains one of the independent Poisson spike trains,
        either n SpikeTrains of the same rate, or len(rate) SpikeTrains with
        varying rates according to the rate parameter. The time unit of the
        SpikeTrains is given by t_stop.
    """
    # Check that the provided input is Hertz
    if not isinstance(rate, pq.Quantity):
        raise ValueError('rate must be a pq.Quantity')

    # Set number n of output spike trains (specified or set to len(rate))
    if not (isinstance(n_spiketrains, int) and n_spiketrains > 0):
        raise ValueError(
            f'n_spiketrains (={n_spiketrains}) must be a positive integer')

    # one rate for all spike trains
    if rate.ndim == 0:
        return StationaryPoissonProcess(
            rate=rate, t_start=t_start, t_stop=t_stop
        ).generate_n_spiketrains(n_spiketrains)

    # different rate for each spike train
    return [StationaryPoissonProcess(
        rate=single_rate, t_start=t_start, t_stop=t_stop).generate_spiketrain()
            for single_rate in rate]


@deprecated_alias(rate_c='coincidence_rate', n='n_spiketrains',
                  return_coinc='return_coincidences')
def single_interaction_process(
        rate, coincidence_rate, t_stop, n_spiketrains=2, jitter=0 * pq.ms,
        coincidences='deterministic', t_start=0 * pq.ms, min_delay=0 * pq.ms,
        return_coincidences=False):
    """
    Generates a multidimensional Poisson SIP (single interaction process)
    plus independent Poisson processes :cite:`generation-Kuhn2003_67`.

    A Poisson SIP consists of Poisson time series which are independent
    except for simultaneous events in all of them. This routine generates
    a SIP plus additional parallel independent Poisson processes.

    Parameters
    ----------
    t_stop : pq.Quantity
        Total time of the simulated processes. The events are drawn between
        0 and `t_stop`.
    rate : pq.Quantity
        Overall mean rate of the time series to be generated (coincidence
        rate `coincidence_rate` is subtracted to determine the background
        rate). Can be:

        * a float, representing the overall mean rate of each process. If
          so, it must be higher than `coincidence_rate`.
        * an iterable of floats (one float per process), each float
          representing the overall mean rate of a process. If so, all the
          entries must be larger than `coincidence_rate`.

    coincidence_rate : pq.Quantity
        Coincidence rate (rate of coincidences for the n-dimensional SIP).
        The SIP spike trains will have coincident events with rate
        `coincidence_rate` plus independent 'background' events with rate
        `rate-rate_coincidence`.
    n_spiketrains : int, optional
        If `rate` is a single pq.Quantity value, `n_spiketrains` specifies the
        number of SpikeTrains to be generated. If rate is an array,
        `n_spiketrains` is ignored and the number of SpikeTrains is equal to
        `len(rate)`.
        Default: 2
    jitter : pq.Quantity, optional
        Jitter for the coincident events. If `jitter == 0`, the events of all
        n correlated processes are exactly coincident. Otherwise, they are
        jittered around a common time randomly, up to +/- `jitter`.
        Default: 0 * pq.ms
    coincidences : {'deterministic', 'stochastic'}, optional
        Whether the total number of injected coincidences must be determin-
        istic (i.e. rate_coincidence is the actual rate with which coincidences
        are generated) or stochastic (i.e. rate_coincidence is the mean rate of
        coincidences):

          * 'deterministic': deterministic rate
          * 'stochastic': stochastic rate

        Default: 'deterministic'
    t_start : pq.Quantity, optional
        Starting time of the series. If specified, it must be lower than
        `t_stop`.
        Default: 0 * pq.ms
    min_delay : pq.Quantity, optional
        Minimum delay between consecutive coincidence times.
        Default: 0 * pq.ms
    return_coincidences : bool, optional
        Whether to return the coincidence times for the SIP process
        Default: False

    Returns
    -------
    output : list
        Realization of a SIP consisting of `n_spiketrains` Poisson processes
        characterized by synchronous events (with the given jitter).
        If `return_coinc` is `True`, the coincidence times are returned as a
        second output argument. They also have an associated time unit (same
        as `t_stop`).

    Examples
    --------
    >>> import quantities as pq
    >>> import elephant.spike_train_generation as stg
    >>> sip, coinc = stg.single_interaction_process(
    ... rate=20*pq.Hz, coincidence_rate=4.*pq.Hz,
    ... t_stop=1*pq.s, n_spiketrains=10, return_coincidences = True)
    """

    # Check if n is a positive integer
    if not (isinstance(n_spiketrains, int) and n_spiketrains > 0):
        raise ValueError(
            f'n_spiketrains (={n_spiketrains}) must be a positive integer')
    if coincidences not in ('deterministic', 'stochastic'):
        raise ValueError(
            "coincidences must be 'deterministic' or 'stochastic'")

    # Assign time unit to jitter, or check that its existing unit is a time
    # unit
    jitter = abs(jitter)

    # Define the array of rates from input argument rate. Check that its length
    # matches with n
    if rate.ndim == 0:
        if rate < 0 * pq.Hz:
            raise ValueError(
                f'rate (={rate}) must be non-negative.')
        rates_b = np.repeat(rate, n_spiketrains)
    else:
        rates_b = rate.flatten()
        if not all(rates_b >= 0.*pq.Hz):
            raise ValueError('*rate* must have non-negative elements')

    # Check: rate>=rate_coincidence
    if np.any(rates_b < coincidence_rate):
        raise ValueError(
            'all elements of *rate* must be >= *rate_coincidence*')

    # Check min_delay < 1./rate_coincidence
    if not (coincidence_rate == 0 * pq.Hz
            or min_delay < 1. / coincidence_rate):
        raise ValueError(
            "'*min_delay* (%s) must be lower than 1/*rate_coincidence* (%s)." %
            (str(min_delay), str((1. / coincidence_rate).rescale(
                min_delay.units))))

    # Generate the n Poisson processes there are the basis for the SIP
    # (coincidences still lacking)
    embedded_poisson_trains = _n_poisson(
        rate=rates_b - coincidence_rate, t_stop=t_stop, t_start=t_start)
    # Convert the trains from neo SpikeTrain objects to simpler pq.Quantity
    # objects
    embedded_poisson_trains = [
        emb.view(pq.Quantity) for emb in embedded_poisson_trains]

    # Generate the array of times for coincident events in SIP, not closer than
    # min_delay. The array is generated as a pq.Quantity.
    if coincidences == 'deterministic':
        # P. Bouss: we want the closest approximation to the average
        # coincidence count.
        n_coincidences = (t_stop - t_start) * coincidence_rate
        # Conversion to integer necessary for python 2
        n_coincidences = int(round(n_coincidences.simplified.item()))
        while True:
            coinc_times = t_start + \
                np.sort(np.random.random(n_coincidences)) * (
                    t_stop - t_start)
            if len(coinc_times) < 2 or min(np.diff(coinc_times)) >= min_delay:
                break
    else:  # coincidences == 'stochastic'
        poisson_process = StationaryPoissonProcess(
            rate=coincidence_rate, t_stop=t_stop, t_start=t_start)
        while True:
            coinc_times = poisson_process.generate_spiketrain()
            if len(coinc_times) < 2 or min(np.diff(coinc_times)) >= min_delay:
                break
        coinc_times = coinc_times.simplified
        units = coinc_times.units
        # Set the coincidence times to T-jitter if larger. This ensures that
        # the last jittered spike time is <T
        effective_t_stop = t_stop - jitter
        coinc_times = np.minimum(coinc_times.magnitude,
                                 effective_t_stop.simplified.magnitude)
        coinc_times = coinc_times * units

    # Replicate coinc_times n times, and jitter each event in each array by
    # +/- jitter (within (t_start, t_stop))
    embedded_coinc = coinc_times + \
        np.random.random(
            (len(rates_b), len(coinc_times))) * 2 * jitter - jitter
    embedded_coinc = embedded_coinc + \
        (t_start - embedded_coinc) * (embedded_coinc < t_start) - \
        (t_stop - embedded_coinc) * (embedded_coinc > t_stop)

    # Inject coincident events into the n SIP processes generated above, and
    # merge with the n independent processes
    sip_process = [
        np.sort(np.concatenate((
            embedded_poisson_trains[m].rescale(t_stop.units),
            embedded_coinc[m].rescale(t_stop.units))) * t_stop.units)
        for m in range(len(rates_b))]

    # Convert back sip_process and coinc_times from pq.Quantity objects to
    # neo.SpikeTrain objects
    sip_process = [
        neo.SpikeTrain(t, t_start=t_start, t_stop=t_stop).rescale(t_stop.units)
        for t in sip_process]
    coinc_times = [
        neo.SpikeTrain(t, t_start=t_start, t_stop=t_stop).rescale(t_stop.units)
        for t in embedded_coinc]

    # Return the processes in the specified output_format
    if not return_coincidences:
        output = sip_process
    else:
        output = sip_process, coinc_times

    return output


def _pool_two_spiketrains(spiketrain_1, spiketrain_2, extremes='inner'):
    """
    Pool the spikes of two spike trains a and b into a unique spike train.

    Parameters
    ----------
    spiketrain_1, spiketrain_2 : neo.SpikeTrain
        Spiketrains to be pooled.
    extremes : {'inner', 'outer'}, optional
        Only spikes of a and b in the specified extremes are considered.
        * 'inner': pool all spikes from max(a.tstart_ b.t_start) to
           min(a.t_stop, b.t_stop)
        * 'outer': pool all spikes from min(a.tstart_ b.t_start) to
           max(a.t_stop, b.t_stop)
        Default: 'inner'

    Returns
    -------
    neo.SpikeTrain
        containing all spikes of the two input spiketrains falling in the
        specified extremes
    """

    unit = spiketrain_1.units
    spiketrain_2 = spiketrain_2.rescale(unit)
    times_1_dimless = spiketrain_1.magnitude
    times_2_dimless = spiketrain_2.rescale(unit).magnitude
    times = np.sort(np.concatenate((times_1_dimless, times_2_dimless)))

    if extremes == 'outer':
        t_start = min(spiketrain_1.t_start, spiketrain_2.t_start)
        t_stop = max(spiketrain_1.t_stop, spiketrain_2.t_stop)
    elif extremes == 'inner':
        t_start = max(spiketrain_1.t_start, spiketrain_2.t_start)
        t_stop = min(spiketrain_1.t_stop, spiketrain_2.t_stop)
        times = times[times > t_start.magnitude]
        times = times[times < t_stop.magnitude]
    else:
        raise ValueError(
            'extremes (%s) can only be "inner" or "outer"' % extremes)

    return neo.SpikeTrain(times=times, units=unit, t_start=t_start,
                          t_stop=t_stop)


def _sample_int_from_pdf(probability_density, n_samples):
    """
    Draw n independent samples from the set {0,1,...,L}, where L=len(a)-1,
    according to the probability distribution a.
    a[j] is the probability to sample j, for each j from 0 to L.


    Parameters
    ----------
    probability_density : np.ndarray
        Probability vector (i..e array of sum 1) that at each entry j carries
        the probability to sample j (j=0,1,...,len(a)-1).
    n_samples : int
        Number of samples generated with the function

    Returns
    -------
    np.ndarray
        An array of n samples taking values between `0` and `n=len(a)-1`.
    """

    cumulative_distribution = np.cumsum(probability_density)
    random_uniforms = np.random.uniform(0, 1, size=n_samples)
    random_uniforms = np.repeat(np.expand_dims(random_uniforms, axis=1),
                                repeats=len(probability_density),
                                axis=1)
    return (cumulative_distribution < random_uniforms).sum(axis=1)


def _mother_proc_cpp_stat(
        amplitude_distribution, t_stop, rate, t_start=0 * pq.ms):
    """
    Generate the hidden ("mother") Poisson process for a Compound Poisson
    Process (CPP).


    Parameters
    ----------
    amplitude_distribution : np.ndarray
        CPP's amplitude distribution :math:`A`. `A[j]` represents the
        probability of a synchronous event of size `j` among the generated
        spike trains. The sum over all entries of :math:`A` must be equal to
        one.
    t_stop : pq.Quantity
        The stopping time of the mother process
    rate : pq.Quantity
        Homogeneous rate of the n spike trains that will be generated by the
        CPP function
    t_start : pq.Quantity, optional
        The starting time of the mother process
        Default: 0 pq.ms

    Returns
    -------
    Poisson spike train representing the mother process generating the CPP
    """
    n_spiketrains = len(amplitude_distribution) - 1
    # expected amplitude
    exp_amplitude = np.dot(
        amplitude_distribution, np.arange(n_spiketrains + 1))
    # expected rate of the mother process
    exp_mother_rate = (n_spiketrains * rate) / exp_amplitude
    return StationaryPoissonProcess(
        rate=exp_mother_rate, t_stop=t_stop, t_start=t_start
    ).generate_spiketrain()


def _cpp_hom_stat(amplitude_distribution, t_stop, rate, t_start=0 * pq.ms):
    """
    Generate a Compound Poisson Process (CPP) with amplitude distribution
    A and heterogeneous firing rates r=r[0], r[1], ..., r[-1].

    Parameters
    ----------
    amplitude_distribution : np.ndarray
        CPP's amplitude distribution :math:`A`. `A[j]` represents the
        probability of a synchronous event of size `j` among the generated
        spike trains. The sum over all entries of :math:`A` must be equal to
        one.
    t_stop : pq.Quantity
        The end time of the output spike trains
    rate : pq.Quantity
        Average rate of each spike train generated
    t_start : pq.Quantity, optional
        The start time of the output spike trains
        Default: 0 pq.ms

    Returns
    -------
    list of neo.SpikeTrain
        with n elements, having average firing rate r and correlated such to
        form a CPP with amplitude distribution a
    """

    # Generate mother process and associated spike labels
    mother = _mother_proc_cpp_stat(
        amplitude_distribution=amplitude_distribution,
        t_stop=t_stop, rate=rate, t_start=t_start)
    labels = _sample_int_from_pdf(amplitude_distribution, len(mother))
    n_spiketrains = len(amplitude_distribution) - 1
    # Number of trains in output

    spiketrains = [[]] * n_spiketrains
    try:  # Faster but more memory-consuming approach
        n_mother_trains = len(mother)  # number of spikes in the mother process
        spike_matrix = np.zeros((n_spiketrains, n_mother_trains), dtype=bool)
        # for each spike, take its label
        for spike_id, label in enumerate(labels):
            # choose label random trains
            train_ids = np.random.choice(n_spiketrains, label, replace=False)
            # and set the spike matrix for that train
            for train_id in train_ids:
                spike_matrix[train_id, spike_id] = True  # and spike to True

        for train_id, row in enumerate(spike_matrix):
            spiketrains[train_id] = mother[row].view(pq.Quantity)

    except MemoryError:  # Slower (~2x) but less memory-consuming approach
        print('memory case')
        for mother_spiketrain, label in zip(mother, labels):
            train_ids = np.random.choice(n_spiketrains, label)
            for train_id in train_ids:
                spiketrains[train_id].append(mother_spiketrain)

    return [neo.SpikeTrain(times=spiketrain, t_start=t_start, t_stop=t_stop)
            for spiketrain in spiketrains]


def _cpp_het_stat(amplitude_distribution, t_stop, rates, t_start=0.*pq.ms):
    """
    Generate a Compound Poisson Process (CPP) with amplitude distribution
    A and heterogeneous firing rates r=r[0], r[1], ..., r[-1].

    Parameters
    ----------
    amplitude_distribution : np.ndarray
        CPP's amplitude distribution :math:`A`. `A[j]` represents the
        probability of a synchronous event of size `j` among the generated
        spike trains. The sum over all entries of :math:`A` must be equal to
        one.
    t_stop : pq.Quantity
        The end time of the output spike trains
    rates : pq.Quantity
        Array of firing rates of each spike train generated with
    t_start : pq.Quantity, optional
        The start time of the output spike trains
        Default: 0 pq.ms

    Returns
    -------
    list of neo.SpikeTrain
        List of neo.SpikeTrains with different firing rates, forming
        a CPP with amplitude distribution `A`.
    """

    # Computation of Parameters of the two CPPs that will be merged
    # (uncorrelated with heterog. rates + correlated with homog. rates)
    n_spiketrains = len(rates)  # number of output spike trains
    # amplitude expectation
    expected_amplitude = np.dot(
        amplitude_distribution, np.arange(n_spiketrains + 1))
    r_sum = np.sum(rates)  # sum of all output firing rates
    r_min = np.min(rates)  # minimum of the firing rates

    # rate of the uncorrelated CPP
    r_uncorrelated = r_sum - n_spiketrains * r_min
    # rate of the correlated CPP
    r_correlated = r_sum / expected_amplitude - r_uncorrelated
    # rate of the hidden mother process
    r_mother = r_uncorrelated + r_correlated

    # Check the analytical constraint for the amplitude distribution
    if amplitude_distribution[1] < (r_uncorrelated / r_mother).rescale(
            pq.dimensionless).magnitude:
        raise ValueError('A[1] too small / A[i], i>1 too high')

    # Compute the amplitude distribution of the correlated CPP, and generate it
    amplitude_distribution = \
        amplitude_distribution * (r_mother / r_correlated).magnitude
    amplitude_distribution[1] = \
        amplitude_distribution[1] - r_uncorrelated / r_correlated
    compound_poisson_spiketrains = _cpp_hom_stat(
        amplitude_distribution, t_stop, r_min, t_start)

    # Generate the independent heterogeneous Poisson processes
    poisson_spiketrains = \
        [StationaryPoissonProcess(
            rate=rate - r_min, t_start=t_start, t_stop=t_stop
                                  ).generate_spiketrain()
         for rate in rates]

    # Pool the correlated CPP and the corresponding Poisson processes
    return [_pool_two_spiketrains(compound_poisson_spiketrain,
                                  poisson_spiketrain)
            for compound_poisson_spiketrain, poisson_spiketrain
            in zip(compound_poisson_spiketrains, poisson_spiketrains)]


@deprecated_alias(A='amplitude_distribution')
def compound_poisson_process(
        rate, amplitude_distribution, t_stop, shift=None, t_start=0 * pq.ms):
    """
    Generate a Compound Poisson Process (CPP; see
    :cite:`generation-Staude2010_327`) with a given `amplitude_distribution`
    :math:`A` and stationary marginal rates `rate`.

    The CPP process is a model for parallel, correlated processes with Poisson
    spiking statistics at pre-defined firing rates. It is composed of
    `len(A)-1` spike trains with a correlation structure determined by the
    amplitude distribution :math:`A`: A[j] is the probability that a spike
    occurs synchronously in any `j` spike trains.

    The CPP is generated by creating a hidden mother Poisson process, and then
    copying spikes of the mother process to `j` of the output spike trains with
    probability `A[j]`.

    Note that this function decorrelates the firing rate of each SpikeTrain
    from the probability for that SpikeTrain to participate in a synchronous
    event (which is uniform across SpikeTrains).

    Parameters
    ----------
    rate : pq.Quantity
        Average rate of each spike train generated. Can be:
          - a single value, all spike trains will have same rate rate
          - an array of values (of length `len(A)-1`), each indicating the
            firing rate of one process in output
    amplitude_distribution : np.ndarray or list
        CPP's amplitude distribution :math:`A`. `A[j]` represents the
        probability of a synchronous event of size `j` among the generated
        spike trains. The sum over all entries of :math:`A` must be equal to
        one.
    t_stop : pq.Quantity
        The end time of the output spike trains.
    shift : pq.Quantity, optional
        If `None`, the injected synchrony is exact.
        If shift is a `pq.Quantity`, all the spike trains are shifted
        independently by a random amount in the interval `[-shift, +shift]`.
        Default: None
    t_start : pq.Quantity, optional
        The `t_start` time of the output spike trains.
        Default: 0 pq.ms

    Returns
    -------
    list of neo.SpikeTrain
        A list of `len(A) - 1` neo.SpikeTrains with specified firing rates
        forming the CPP with amplitude distribution :math:`A`.
    """
    if not isinstance(amplitude_distribution, np.ndarray):
        amplitude_distribution = np.array(amplitude_distribution)
    # Check A is a probability distribution (it sums to 1 and is positive)
    if abs(sum(amplitude_distribution) - 1) > np.finfo('float').eps:
        raise ValueError(
            f"'amplitude_distribution' must be a probability vector: "
            f"sum(A) = {sum(amplitude_distribution)} != 1")
    if np.any(amplitude_distribution < 0):
        raise ValueError("'amplitude_distribution' must be a probability "
                         "vector with positive entries")
    # Check that the rate is not an empty pq.Quantity
    if rate.ndim == 1 and len(rate) == 0:
        raise ValueError('Rate is an empty pq.Quantity array')
    # Return empty spike trains for specific parameters
    if amplitude_distribution[0] == 1 or np.sum(np.abs(rate.magnitude)) == 0:
        return [neo.SpikeTrain([] * t_stop.units,
                               t_stop=t_stop,
                               t_start=t_start)] * (
                len(amplitude_distribution) - 1)

    # Homogeneous rates
    if rate.ndim == 0:
        compound_poisson_spiketrains = _cpp_hom_stat(
            amplitude_distribution=amplitude_distribution,
            t_stop=t_stop, rate=rate,
            t_start=t_start)
    # Heterogeneous rates
    else:
        compound_poisson_spiketrains = _cpp_het_stat(
            amplitude_distribution=amplitude_distribution,
            t_stop=t_stop, rates=rate,
            t_start=t_start)

    if shift is not None:
        # Dither the output spiketrains
        compound_poisson_spiketrains = \
            [dither_spike_train(spiketrain, shift=shift, edges=True)[0]
             for spiketrain in compound_poisson_spiketrains]

    return compound_poisson_spiketrains


# Alias for :func:`compound_poisson_process`
cpp = compound_poisson_process



# -*- coding: utf-8 -*-
"""
GPFA core functionality.

:copyright: Copyright 2014-2023 by the Elephant team, see AUTHORS.txt.
:license: Modified BSD, see LICENSE.txt for details.
"""

from __future__ import division, print_function, unicode_literals

import time
import warnings

import numpy as np
import scipy.linalg as linalg
import scipy.optimize as optimize
import scipy.sparse as sparse
from sklearn.decomposition import FactorAnalysis
from tqdm import trange

from . import gpfa_util


def fit(seqs_train, x_dim=3, bin_width=20.0, min_var_frac=0.01, em_tol=1.0E-8,
        em_max_iters=500, tau_init=100.0, eps_init=1.0E-3, freq_ll=5,
        verbose=False):
    """
    Fit the GPFA model with the given training data.

    Parameters
    ----------
    seqs_train : np.recarray
        training data structure, whose n-th element (corresponding to
        the n-th experimental trial) has fields
        T : int
            number of bins
        y : (#units, T) np.ndarray
            neural data
    x_dim : int, optional
        state dimensionality
        Default: 3
    bin_width : float, optional
        spike bin width in msec
        Default: 20.0
    min_var_frac : float, optional
        fraction of overall data variance for each observed dimension to set as
        the private variance floor.  This is used to combat Heywood cases,
        where ML parameter learning returns one or more zero private variances.
        Default: 0.01
        (See Martin & McDonald, Psychometrika, Dec 1975.)
    em_tol : float, optional
        stopping criterion for EM
        Default: 1e-8
    em_max_iters : int, optional
        number of EM iterations to run
        Default: 500
    tau_init : float, optional
        GP timescale initialization in msec
        Default: 100
    eps_init : float, optional
        GP noise variance initialization
        Default: 1e-3
    freq_ll : int, optional
        data likelihood is computed at every freq_ll EM iterations. freq_ll = 1
        means that data likelihood is computed at every iteration.
        Default: 5
    verbose : bool, optional
        specifies whether to display status messages
        Default: False

    Returns
    -------
    parameter_estimates : dict
        Estimated model parameters.
        When the GPFA method is used, following parameters are contained
            covType: {'rbf', 'tri', 'logexp'}
                type of GP covariance
            gamma: np.ndarray of shape (1, #latent_vars)
                related to GP timescales by 'bin_width / sqrt(gamma)'
            eps: np.ndarray of shape (1, #latent_vars)
                GP noise variances
            d: np.ndarray of shape (#units, 1)
                observation mean
            C: np.ndarray of shape (#units, #latent_vars)
                mapping between the neuronal data space and the latent variable
                space
            R: np.ndarray of shape (#units, #latent_vars)
                observation noise covariance

    fit_info : dict
        Information of the fitting process and the parameters used there
        iteration_time : list
            containing the runtime for each iteration step in the EM algorithm.
    """
    # For compute efficiency, train on equal-length segments of trials
    seqs_train_cut = gpfa_util.cut_trials(seqs_train)
    if len(seqs_train_cut) == 0:
        warnings.warn('No segments extracted for training. Defaulting to '
                      'segLength=Inf.')
        seqs_train_cut = gpfa_util.cut_trials(seqs_train, seg_length=np.inf)

    # ==================================
    # Initialize state model parameters
    # ==================================
    params_init = dict()
    params_init['covType'] = 'rbf'
    # GP timescale
    # Assume binWidth is the time step size.
    params_init['gamma'] = (bin_width / tau_init) ** 2 * np.ones(x_dim)
    # GP noise variance
    params_init['eps'] = eps_init * np.ones(x_dim)

    # ========================================
    # Initialize observation model parameters
    # ========================================
    print('Initializing parameters using factor analysis...')

    y_all = np.hstack(seqs_train_cut['y'])
    fa = FactorAnalysis(n_components=x_dim, copy=True,
                        noise_variance_init=np.diag(np.cov(y_all, bias=True)))
    fa.fit(y_all.T)
    params_init['d'] = y_all.mean(axis=1)
    params_init['C'] = fa.components_.T
    params_init['R'] = np.diag(fa.noise_variance_)

    # Define parameter constraints
    params_init['notes'] = {
        'learnKernelParams': True,
        'learnGPNoise': False,
        'RforceDiagonal': True,
    }

    # =====================
    # Fit model parameters
    # =====================
    print('\nFitting GPFA model...')

    params_est, seqs_train_cut, ll_cut, iter_time = em(
        params_init, seqs_train_cut, min_var_frac=min_var_frac,
        max_iters=em_max_iters, tol=em_tol, freq_ll=freq_ll, verbose=verbose)

    fit_info = {'iteration_time': iter_time, 'log_likelihoods': ll_cut}

    return params_est, fit_info


def em(params_init, seqs_train, max_iters=500, tol=1.0E-8, min_var_frac=0.01,
       freq_ll=5, verbose=False):
    """
    Fits GPFA model parameters using expectation-maximization (EM) algorithm.

    Parameters
    ----------
    params_init : dict
        GPFA model parameters at which EM algorithm is initialized
        covType : {'rbf', 'tri', 'logexp'}
            type of GP covariance
        gamma : np.ndarray of shape (1, #latent_vars)
            related to GP timescales by
            'bin_width / sqrt(gamma)'
        eps : np.ndarray of shape (1, #latent_vars)
            GP noise variances
        d : np.ndarray of shape (#units, 1)
            observation mean
        C : np.ndarray of shape (#units, #latent_vars)
            mapping between the neuronal data space and the
            latent variable space
        R : np.ndarray of shape (#units, #latent_vars)
            observation noise covariance
    seqs_train : np.recarray
        training data structure, whose n-th entry (corresponding to the n-th
        experimental trial) has fields
        T : int
            number of bins
        y : np.ndarray (yDim x T)
            neural data
    max_iters : int, optional
        number of EM iterations to run
        Default: 500
    tol : float, optional
        stopping criterion for EM
        Default: 1e-8
    min_var_frac : float, optional
        fraction of overall data variance for each observed dimension to set as
        the private variance floor.  This is used to combat Heywood cases,
        where ML parameter learning returns one or more zero private variances.
        Default: 0.01
        (See Martin & McDonald, Psychometrika, Dec 1975.)
    freq_ll : int, optional
        data likelihood is computed at every freq_ll EM iterations.
        freq_ll = 1 means that data likelihood is computed at every
        iteration.
        Default: 5
    verbose : bool, optional
        specifies whether to display status messages
        Default: False

    Returns
    -------
    params_est : dict
        GPFA model parameter estimates, returned by EM algorithm (same
        format as params_init)
    seqs_latent : np.recarray
        a copy of the training data structure, augmented with the new
        fields:
        latent_variable : np.ndarray of shape (#latent_vars x #bins)
            posterior mean of latent variables at each time bin
        Vsm : np.ndarray of shape (#latent_vars, #latent_vars, #bins)
            posterior covariance between latent variables at each
            timepoint
        VsmGP : np.ndarray of shape (#bins, #bins, #latent_vars)
            posterior covariance over time for each latent
            variable
    ll : list
        list of log likelihoods after each EM iteration
    iter_time : list
        lisf of computation times (in seconds) for each EM iteration
    """
    params = params_init
    t = seqs_train['T']
    y_dim, x_dim = params['C'].shape
    lls = []
    ll_old = ll_base = ll = 0.0
    iter_time = []
    var_floor = min_var_frac * np.diag(np.cov(np.hstack(seqs_train['y'])))
    seqs_latent = None

    # Loop once for each iteration of EM algorithm
    for iter_id in trange(1, max_iters + 1, desc='EM iteration',
                          disable=not verbose):
        if verbose:
            print()
        tic = time.time()
        get_ll = (np.fmod(iter_id, freq_ll) == 0) or (iter_id <= 2)

        # ==== E STEP =====
        if not np.isnan(ll):
            ll_old = ll
        seqs_latent, ll = exact_inference_with_ll(seqs_train, params,
                                                  get_ll=get_ll)
        lls.append(ll)

        # ==== M STEP ====
        sum_p_auto = np.zeros((x_dim, x_dim))
        for seq_latent in seqs_latent:
            sum_p_auto += seq_latent['Vsm'].sum(axis=2) \
                + seq_latent['latent_variable'].dot(
                seq_latent['latent_variable'].T)
        y = np.hstack(seqs_train['y'])
        latent_variable = np.hstack(seqs_latent['latent_variable'])
        sum_yxtrans = y.dot(latent_variable.T)
        sum_xall = latent_variable.sum(axis=1)[:, np.newaxis]
        sum_yall = y.sum(axis=1)[:, np.newaxis]

        # term is (xDim+1) x (xDim+1)
        term = np.vstack([np.hstack([sum_p_auto, sum_xall]),
                          np.hstack([sum_xall.T, t.sum().reshape((1, 1))])])
        # yDim x (xDim+1)
        cd = gpfa_util.rdiv(np.hstack([sum_yxtrans, sum_yall]), term)

        params['C'] = cd[:, :x_dim]
        params['d'] = cd[:, -1]

        # yCent must be based on the new d
        # yCent = bsxfun(@minus, [seq.y], currentParams.d);
        # R = (yCent * yCent' - (yCent * [seq.latent_variable]') * \
        #     currentParams.C') / sum(T);
        c = params['C']
        d = params['d'][:, np.newaxis]
        if params['notes']['RforceDiagonal']:
            sum_yytrans = (y * y).sum(axis=1)[:, np.newaxis]
            yd = sum_yall * d
            term = ((sum_yxtrans - d.dot(sum_xall.T)) * c).sum(axis=1)
            term = term[:, np.newaxis]
            r = d ** 2 + (sum_yytrans - 2 * yd - term) / t.sum()

            # Set minimum private variance
            r = np.maximum(var_floor, r)
            params['R'] = np.diag(r[:, 0])
        else:
            sum_yytrans = y.dot(y.T)
            yd = sum_yall.dot(d.T)
            term = (sum_yxtrans - d.dot(sum_xall.T)).dot(c.T)
            r = d.dot(d.T) + (sum_yytrans - yd - yd.T - term) / t.sum()

            params['R'] = (r + r.T) / 2  # ensure symmetry

        if params['notes']['learnKernelParams']:
            res = learn_gp_params(seqs_latent, params, verbose=verbose)
            params['gamma'] = res['gamma']

        t_end = time.time() - tic
        iter_time.append(t_end)

        # Verify that likelihood is growing monotonically
        if iter_id <= 2:
            ll_base = ll
        elif verbose and ll < ll_old:
            print('\nError: Data likelihood has decreased ',
                  'from {0} to {1}'.format(ll_old, ll))
        elif (ll - ll_base) < (1 + tol) * (ll_old - ll_base):
            break

    if len(lls) < max_iters:
        print('Fitting has converged after {0} EM iterations.)'.format(
            len(lls)))

    if np.any(np.diag(params['R']) == var_floor):
        warnings.warn('Private variance floor used for one or more observed '
                      'dimensions in GPFA.')

    return params, seqs_latent, lls, iter_time


def exact_inference_with_ll(seqs, params, get_ll=True):
    """
    Extracts latent trajectories from neural data, given GPFA model parameters.

    Parameters
    ----------
    seqs : np.recarray
        Input data structure, whose n-th element (corresponding to the n-th
        experimental trial) has fields:
        y : np.ndarray of shape (#units, #bins)
            neural data
        T : int
            number of bins
    params : dict
        GPFA model parameters whe the following fields:
        C : np.ndarray
            FA factor loadings matrix
        d : np.ndarray
            FA mean vector
        R : np.ndarray
            FA noise covariance matrix
        gamma : np.ndarray
            GP timescale
        eps : np.ndarray
            GP noise variance
    get_ll : bool, optional
          specifies whether to compute data log likelihood (default: True)

    Returns
    -------
    seqs_latent : np.recarray
        a copy of the input data structure, augmented with the new
        fields:
        latent_variable :  (#latent_vars, #bins) np.ndarray
              posterior mean of latent variables at each time bin
        Vsm :  (#latent_vars, #latent_vars, #bins) np.ndarray
              posterior covariance between latent variables at each
              timepoint
        VsmGP :  (#bins, #bins, #latent_vars) np.ndarray
                posterior covariance over time for each latent
                variable
    ll : float
        data log likelihood, np.nan is returned when `get_ll` is set False
    """
    y_dim, x_dim = params['C'].shape

    # copy the contents of the input data structure to output structure
    dtype_out = [(x, seqs[x].dtype) for x in seqs.dtype.names]
    dtype_out.extend([('latent_variable', object), ('Vsm', object),
                      ('VsmGP', object)])
    seqs_latent = np.empty(len(seqs), dtype=dtype_out)
    for dtype_name in seqs.dtype.names:
        seqs_latent[dtype_name] = seqs[dtype_name]

    # Precomputations
    if params['notes']['RforceDiagonal']:
        rinv = np.diag(1.0 / np.diag(params['R']))
        logdet_r = (np.log(np.diag(params['R']))).sum()
    else:
        rinv = linalg.inv(params['R'])
        rinv = (rinv + rinv.T) / 2  # ensure symmetry
        logdet_r = gpfa_util.logdet(params['R'])

    c_rinv = params['C'].T.dot(rinv)
    c_rinv_c = c_rinv.dot(params['C'])

    t_all = seqs_latent['T']
    t_uniq = np.unique(t_all)
    ll = 0.

    # Overview:
    # - Outer loop on each element of Tu.
    # - For each element of Tu, find all trials with that length.
    # - Do inference and LL computation for all those trials together.
    for t in t_uniq:
        k_big, k_big_inv, logdet_k_big = gpfa_util.make_k_big(params, t)
        k_big = sparse.csr_matrix(k_big)

        blah = [c_rinv_c for _ in range(t)]
        c_rinv_c_big = linalg.block_diag(*blah)  # (xDim*T) x (xDim*T)
        minv, logdet_m = gpfa_util.inv_persymm(k_big_inv + c_rinv_c_big, x_dim)

        # Note that posterior covariance does not depend on observations,
        # so can compute once for all trials with same T.
        # xDim x xDim posterior covariance for each timepoint
        vsm = np.full((x_dim, x_dim, t), np.nan)
        idx = np.arange(0, x_dim * t + 1, x_dim)
        for i in range(t):
            vsm[:, :, i] = minv[idx[i]:idx[i + 1], idx[i]:idx[i + 1]]

        # T x T posterior covariance for each GP
        vsm_gp = np.full((t, t, x_dim), np.nan)
        for i in range(x_dim):
            vsm_gp[:, :, i] = minv[i::x_dim, i::x_dim]

        # Process all trials with length T
        n_list = np.where(t_all == t)[0]
        # dif is yDim x sum(T)
        dif = np.hstack(seqs_latent[n_list]['y']) - params['d'][:, np.newaxis]
        # term1Mat is (xDim*T) x length(nList)
        term1_mat = c_rinv.dot(dif).reshape((x_dim * t, -1), order='F')

        # Compute blkProd = CRinvC_big * invM efficiently
        # blkProd is block persymmetric, so just compute top half
        t_half = int(np.ceil(t / 2.0))
        blk_prod = np.zeros((x_dim * t_half, x_dim * t))
        idx = range(0, x_dim * t_half + 1, x_dim)
        for i in range(t_half):
            blk_prod[idx[i]:idx[i + 1], :] = c_rinv_c.dot(
                minv[idx[i]:idx[i + 1], :])
        blk_prod = k_big[:x_dim * t_half, :].dot(
            gpfa_util.fill_persymm(np.eye(x_dim * t_half, x_dim * t) -
                                   blk_prod, x_dim, t))
        # latent_variableMat is (xDim*T) x length(nList)
        latent_variable_mat = gpfa_util.fill_persymm(
            blk_prod, x_dim, t).dot(term1_mat)

        for i, n in enumerate(n_list):
            seqs_latent[n]['latent_variable'] = \
                latent_variable_mat[:, i].reshape((x_dim, t), order='F')
            seqs_latent[n]['Vsm'] = vsm
            seqs_latent[n]['VsmGP'] = vsm_gp

        if get_ll:
            # Compute data likelihood
            val = -t * logdet_r - logdet_k_big - logdet_m \
                  - y_dim * t * np.log(2 * np.pi)
            ll = ll + len(n_list) * val - (rinv.dot(dif) * dif).sum() \
                + (term1_mat.T.dot(minv) * term1_mat.T).sum()

    if get_ll:
        ll /= 2
    else:
        ll = np.nan

    return seqs_latent, ll


def learn_gp_params(seqs_latent, params, verbose=False):
    """Updates parameters of GP state model, given neural trajectories.

    Parameters
    ----------
    seqs_latent : np.recarray
        data structure containing neural trajectories;
    params : dict
        current GP state model parameters, which gives starting point
        for gradient optimization;
    verbose : bool, optional
        specifies whether to display status messages (default: False)

    Returns
    -------
    param_opt : np.ndarray
        updated GP state model parameter

    Raises
    ------
    ValueError
        If `params['covType'] != 'rbf'`.
        If `params['notes']['learnGPNoise']` set to True.

    """
    if params['covType'] != 'rbf':
        raise ValueError("Only 'rbf' GP covariance type is supported.")
    if params['notes']['learnGPNoise']:
        raise ValueError("learnGPNoise is not supported.")
    param_name = 'gamma'

    param_init = params[param_name]
    param_opt = {param_name: np.empty_like(param_init)}

    x_dim = param_init.shape[-1]
    precomp = gpfa_util.make_precomp(seqs_latent, x_dim)

    # Loop once for each state dimension (each GP)
    for i in range(x_dim):
        const = {'eps': params['eps'][i]}
        initp = np.log(param_init[i])
        res_opt = optimize.minimize(gpfa_util.grad_betgam, initp,
                                    args=(precomp[i], const),
                                    method='L-BFGS-B', jac=True)
        param_opt['gamma'][i] = np.exp(res_opt.x.item())

        if verbose:
            print('\n Converged p; xDim:{}, p:{}'.format(i, res_opt.x))

    return param_opt


def orthonormalize(params_est, seqs):
    """
    Orthonormalize the columns of the loading matrix C and apply the
    corresponding linear transform to the latent variables.

    Parameters
    ----------
    params_est : dict
        First return value of extract_trajectory() on the training data set.
        Estimated model parameters.
        When the GPFA method is used, following parameters are contained
        covType : {'rbf', 'tri', 'logexp'}
            type of GP covariance
            Currently, only 'rbf' is supported.
        gamma : np.ndarray of shape (1, #latent_vars)
            related to GP timescales by 'bin_width / sqrt(gamma)'
        eps : np.ndarray of shape (1, #latent_vars)
            GP noise variances
        d : np.ndarray of shape (#units, 1)
            observation mean
        C : np.ndarray of shape (#units, #latent_vars)
            mapping between the neuronal data space and the latent variable
            space
        R : np.ndarray of shape (#units, #latent_vars)
            observation noise covariance

    seqs : np.recarray
        Contains the embedding of the training data into the latent variable
        space.
        Data structure, whose n-th entry (corresponding to the n-th
        experimental trial) has fields
        T : int
          number of timesteps
        y : np.ndarray of shape (#units, #bins)
          neural data
        latent_variable : np.ndarray of shape (#latent_vars, #bins)
          posterior mean of latent variables at each time bin
        Vsm : np.ndarray of shape (#latent_vars, #latent_vars, #bins)
          posterior covariance between latent variables at each
          timepoint
        VsmGP : np.ndarray of shape (#bins, #bins, #latent_vars)
          posterior covariance over time for each latent variable

    Returns
    -------
    params_est : dict
        Estimated model parameters, including `Corth`, obtained by
        orthonormalizing the columns of C.
    seqs : np.recarray
        Training data structure that contains the new field
        `latent_variable_orth`, the orthonormalized neural trajectories.
    """
    C = params_est['C']
    X = np.hstack(seqs['latent_variable'])
    latent_variable_orth, Corth, _ = gpfa_util.orthonormalize(X, C)
    seqs = gpfa_util.segment_by_trial(
        seqs, latent_variable_orth, 'latent_variable_orth')

    params_est['Corth'] = Corth

    return Corth, seqs



# -*- coding: utf-8 -*-

"""
The Change point detection algorithm :cite:`cpd-Messer2014_2027` determines if
a spike train `spiketrain` can be considered as a stationary process (constant
firing rate) or not as stationary process (i.e. presence of one or more points
at which the rate increases or decreases). In case of non-stationarity, the
output is a list of detected Change Points (CPs).

Essentially, a set of two-sided windows of width `h`
(`_filter(t, h, spiketrain)`) slides over the spike train within the time
`[h, t_final-h]`. This generates a `_filter_process(time_step, h, spiketrain)`
that assigns at each time `t` the difference between a spike lying in the right
and left windows. If at any time `t` this difference is large 'enough', the
presence of a rate Change Point in a neighborhood of `t` is assumed. A
threshold `test_quantile` for the maximum of the filter_process (max
difference of spike count between the left and right windows) is derived based
on asymptotic considerations. The procedure is repeated for an arbitrary set of
windows with different sizes `h`.

.. autosummary::
    :toctree: _toctree/change_point_detection

    multiple_filter_test
    empirical_parameters

Examples
--------
>>> import quantities as pq
>>> from elephant.change_point_detection import multiple_filter_test
>>> spike_times = [1.1, 1.2, 1.4, 1.6, 1.7, 1.75, 1.8, 1.9, 1.95] * pq.s
>>> change_points = multiple_filter_test(window_sizes=[0.5] * pq.s,
...     spiketrain=spike_times, t_final=2.1 * pq.s, alpha=5, n_surrogates=100,
...     time_step=0.1 * pq.s) # doctest: +SKIP
[[array(1.5) * s]]

Original code
-------------
Adapted from the published R implementation: :cite:`cpd-Messer2014_2027`

"""

from __future__ import division, print_function, unicode_literals

import numpy as np
import quantities as pq

from elephant.utils import deprecated_alias

__all__ = [
    "multiple_filter_test",
    "empirical_parameters"
]


@deprecated_alias(dt='time_step')
def multiple_filter_test(window_sizes, spiketrain, t_final, alpha,
                         n_surrogates=1000, test_quantile=None,
                         test_param=None, time_step=None):
    """
    Detects change points.

    This function returns the detected change points that corresponds to the
    maxima of the *filter processes* - the processes generated by sliding
    windows of step `time_step`; at each step the difference between spikes on
    the right and left windows is calculated.

    Parameters
    ----------
    window_sizes : list of pq.Quantity
                list that contains windows sizes
    spiketrain : neo.SpikeTrain or pq.Quantity
        A spiketrain object to analyze.
    t_final : pq.Quantity
        The final time of the spike train which is to be analysed
    alpha : float
        Alpha-quantile in range [0, 100] for the set of maxima of the limit
        processes
    n_surrogates : int, optional
        The number of simulated limit processes.
        Default: 1000
    test_quantile : float or None, optional
        The threshold for the maxima of the filter derivative processes; if any
        of these maxima is larger than this value, it is assumed the
        presence of a change point (cp) at the time corresponding to that
        maximum.
        If None, will be set according to the :func:`empirical_parameters`.
        Default: None
    test_param : (3, num. of windows) np.ndarray or None, optional
        first row: list of `h`, second and third rows: empirical means and
        variances of the limit process corresponding to `h`. This will be
        used to normalize the *filter processes* in order to give to the
        every maximum the same impact on the global statistic.
        If None, will be set according to the :func:`empirical_parameters`.
        Default: None
    time_step : pq.Quantity or None, optional
        The resolution - the time step at which the windows are slided.
        If None, will be set to ``window_size / 20``.
        Default: None

    Returns
    -------
    cps : list of list
        The change points,
        one list for each window size `h`, containing the points detected with
        the corresponding `filter_process`. N.B.: only cps whose h-neighborhood
        does not include previously detected cps (with smaller window h) are
        added to the list.

    """

    if test_quantile is None and test_param is None:
        test_quantile, test_param = empirical_parameters(window_sizes, t_final,
                                                         alpha, n_surrogates,
                                                         time_step)
    elif test_quantile is None:
        test_quantile = empirical_parameters(window_sizes, t_final, alpha,
                                             n_surrogates, time_step)[0]
    elif test_param is None:
        test_param = empirical_parameters(window_sizes, t_final, alpha,
                                          n_surrogates, time_step)[1]

    #  List of lists of detected change points (CPs), to be returned
    cps = []

    for i, h in enumerate(window_sizes):
        # automatic setting of time_step
        dt_temp = h / 20 if time_step is None else time_step
        # filter_process for window of size h
        t, differences = _filter_process(dt_temp, h, spiketrain, t_final,
                                         test_param)
        time_index = np.arange(len(differences))
        # Point detected with window h
        cps_window = []
        while np.max(differences) > test_quantile:
            cp_index = np.argmax(differences)
            # from index to time
            cp = cp_index * dt_temp + h
            # before repeating the procedure, the h-neighbourgs of detected CP
            # are discarded, because rate changes into it are alrady explained
            mask_fore = time_index > cp_index - int((h / dt_temp).simplified)
            mask_back = time_index < cp_index + int((h / dt_temp).simplified)
            differences[mask_fore & mask_back] = 0
            # check if the neighbourhood of detected cp does not contain cps
            # detected with other windows
            neighbourhood_free = True
            # iterate on lists of cps detected with smaller window
            for j in range(i):
                # iterate on CPs detected with the j-th smallest window
                for c_pre in cps[j]:
                    if c_pre - h < cp < c_pre + h:
                        neighbourhood_free = False
                        break
            # if none of the previously detected CPs falls in the h-
            # neighbourhood
            if neighbourhood_free:
                # add the current CP to the list
                cps_window.append(cp)
        # add the present list to the grand list
        cps.append(cps_window)

    return cps


def _brownian_motion(t_in, t_fin, x_in, time_step):
    """
    Generate a Brownian Motion.

    Parameters
    ----------
    t_in : quantities,
        initial time
    t_fin : quantities,
         final time
    x_in : float,
        initial point of the process: _brownian_motio(0) = x_in
    time_step : quantities,
      resolution, time step at which brownian increments are summed
    Returns
    -------
    Brownian motion on [t_in, t_fin], with resolution time_step and initial
    state x_in
    """

    u = 1 * pq.s
    try:
        t_in_sec = t_in.rescale(u).magnitude
    except ValueError:
        raise ValueError("t_in must be a time quantity")
    try:
        t_fin_sec = t_fin.rescale(u).magnitude
    except ValueError:
        raise ValueError("t_fin must be a time quantity")
    try:
        dt_sec = time_step.rescale(u).magnitude
    except ValueError:
        raise ValueError("dt must be a time quantity")

    x = np.random.normal(0, np.sqrt(dt_sec),
                         size=int((t_fin_sec - t_in_sec) / dt_sec))
    s = np.cumsum(x)
    return s + x_in


def _limit_processes(window_sizes, t_final, time_step):
    """
    Generate the limit processes (depending only on t_final and h), one for
    each window size `h` in H. The distribution of maxima of these processes
    is used to derive threshold `test_quantile` and parameters `test_param`.

    Parameters
    ----------
        window_sizes : list of quantities
            set of windows' size
        t_final : quantity object
            end of limit process
        time_step : quantity object
            resolution, time step at which the windows are slided

    Returns
    -------
        limit_processes : list of numpy array
            each entries contains the limit processes for each h,
            evaluated in [h,T-h] with steps time_step
    """

    limit_processes = []

    u = 1 * pq.s
    try:
        window_sizes_sec = window_sizes.rescale(u).magnitude
    except ValueError:
        raise ValueError("window_sizes must be a list of times")
    try:
        dt_sec = time_step.rescale(u).magnitude
    except ValueError:
        raise ValueError("time_step must be a time quantity")

    w = _brownian_motion(0 * u, t_final, 0, time_step)

    for h in window_sizes_sec:
        # BM on [h,T-h], shifted in time t-->t+h
        brownian_right = w[int(2 * h / dt_sec):]
        # BM on [h,T-h], shifted in time t-->t-h
        brownian_left = w[:int(-2 * h / dt_sec)]
        # BM on [h,T-h]
        brownian_center = w[int(h / dt_sec):int(-h / dt_sec)]

        modul = np.abs(brownian_right + brownian_left - 2 * brownian_center)
        limit_process_h = modul / (np.sqrt(2 * h))
        limit_processes.append(limit_process_h)

    return limit_processes


@deprecated_alias(dt='time_step')
def empirical_parameters(window_sizes, t_final, alpha, n_surrogates=1000,
                         time_step=None):
    r"""
    This function generates the threshold and the null parameters.
    The filter processes (`h`) have been proved to converge (for `t_final`,
    :math:`h \to \infty`) to a continuous functional of a Brownian motion
    ('limit_process'). Using a MonteCarlo technique, maxima of
    these limit_processes are collected.

    The threshold is defined as the alpha quantile of this set of maxima.
    Namely:

    test_quantile := alpha quantile of
    :math:`{\max_{h \in \text{window\_sizes}} \max_{t \in [h, t_{final}-h]}
    \text{limit\_process}_h(t)}`

    Parameters
    ----------
    window_sizes : list of pq.Quantity
                list that contains windows sizes
    t_final : pq.Quantity
        The final time of the spike train which is to be analysed
    alpha : float
        Alpha-quantile in range [0, 100] for the set of maxima of the limit
        processes
    n_surrogates : int, optional
        The number of simulated limit processes.
        Default: 1000
    time_step : pq.Quantity or None, optional
        The resolution - the time step at which the windows are slided.
        If None, will be set to ``window_size / 20``.
        Default: None

    Returns
    -------
    test_quantile : float
        The threshold for the maxima of the filter derivative processes; if any
        of these maxima is larger than this value, it is assumed the
        presence of a change point (cp) at the time corresponding to that
        maximum.
    test_param : (3, num. of windows) np.ndarray
        first row: list of `h`, second and third rows: empirical means and
        variances of the limit process corresponding to `h`. This will be
        used to normalize the *filter processes* in order to give to the
        every maximum the same impact on the global statistic.

    Examples
    --------
    >>> import quantities as pq
    >>> from elephant.change_point_detection import empirical_parameters
    >>> test_quantile, test_param = empirical_parameters(
    ...     window_sizes=[0.5] * pq.s, t_final=2.1 * pq.s, alpha=5,
    ...     n_surrogates=100, time_step=0.1 * pq.s)
    >>> test_quantile  # doctest: +SKIP
    1.8133759165692873
    >>> test_param # doctest: +SKIP
    array([[0.5       ],
           [1.74482974],
           [0.24290945]])
    """

    # try:
    #     window_sizes_sec = window_sizes.rescale(u)
    # except ValueError:
    #     raise ValueError("H must be a list of times")
    # window_sizes_mag = window_sizes_sec.magnitude
    # try:
    #     t_final_sec = t_final.rescale(u)
    # except ValueError:
    #     raise ValueError("T must be a time quantity")
    # t_final_mag = t_final_sec.magnitude

    if not isinstance(window_sizes, pq.Quantity):
        raise ValueError("window_sizes must be a list of time quantities")
    if not isinstance(t_final, pq.Quantity):
        raise ValueError("t_final must be a time quantity")
    if not isinstance(n_surrogates, int):
        raise TypeError("n_surrogates must be an integer")
    if not (isinstance(time_step, pq.Quantity) or (time_step is None)):
        raise ValueError("time_step must be a time quantity")

    if t_final <= 0:
        raise ValueError("t_final needs to be strictly positive")
    if alpha * (100.0 - alpha) < 0:
        raise ValueError("alpha needs to be in (0,100)")
    if np.min(window_sizes) <= 0:
        raise ValueError("window size needs to be strictly positive")
    if np.max(window_sizes) >= t_final / 2:
        raise ValueError("window size too large")
    if time_step is not None:
        for h in window_sizes:
            if int(h.rescale('us')) % int(time_step.rescale('us')) != 0:
                raise ValueError(
                    "Every window size h must be a multiple of time_step")

    # Generate a matrix M*: n X m where n = n_surrogates is the number of
    # simulated limit processes and m is the number of chosen window sizes.
    # Elements are: M*(i,h) = max(t in T)[`limit_process_h`(t)],
    # for each h in H and surrogate i
    maxima_matrix = []

    for i in range(n_surrogates):
        # mh_star = []
        simu = _limit_processes(window_sizes, t_final, time_step)
        # for i, h in enumerate(window_sizes_mag):
        #     # max over time of the limit process generated with window h
        #     m_h = np.max(simu[i])
        #     mh_star.append(m_h)
        # max over time of the limit process generated with window h
        mh_star = [np.max(x) for x in simu]
        maxima_matrix.append(mh_star)

    maxima_matrix = np.asanyarray(maxima_matrix)

    # these parameters will be used to normalize both the limit_processes (H0)
    # and the filter_processes
    null_mean = maxima_matrix.mean(axis=0)
    null_var = maxima_matrix.var(axis=0)

    # matrix normalization by mean and variance of the limit process, in order
    # to give, for every h, the same impact on the global maximum
    matrix_normalized = (maxima_matrix - null_mean) / np.sqrt(null_var)

    great_maxs = np.max(matrix_normalized, axis=1)
    test_quantile = np.percentile(great_maxs, 100.0 - alpha)
    null_parameters = [window_sizes, null_mean, null_var]
    test_param = np.asanyarray(null_parameters)

    return test_quantile, test_param


def _filter(t_center, window, spiketrain):
    """
    This function calculates the difference of spike counts in the left and
    right side of a window of size h centered in t and normalized by its
    variance. The variance of this count can be expressed as a combination of
    mean and var of the I.S.I. lying inside the window.

    Parameters
    ----------
    t_center : quantity
        time on which the window is centered
    window : quantity
        window's size
    spiketrain : list, numpy array or SpikeTrain
        spike train to analyze

    Returns
    -------
    difference : float,
        difference of spike count normalized by its variance
    """

    u = 1 * pq.s
    try:
        t_sec = t_center.rescale(u).magnitude
    except AttributeError:
        raise ValueError("t must be a quantities object")
    # tm = t_sec.magnitude
    try:
        h_sec = window.rescale(u).magnitude
    except AttributeError:
        raise ValueError("h must be a time quantity")
    # hm = h_sec.magnitude
    try:
        spk_sec = spiketrain.rescale(u).magnitude
    except AttributeError:
        raise ValueError(
            "spiketrain must be a list (array) of times or a neo spiketrain")

    # cut spike-train on the right
    train_right = spk_sec[(t_sec < spk_sec) & (spk_sec < t_sec + h_sec)]
    # cut spike-train on the left
    train_left = spk_sec[(t_sec - h_sec < spk_sec) & (spk_sec < t_sec)]
    # spike count in the right side
    count_right = train_right.size
    # spike count in the left side
    count_left = train_left.size
    # form spikes to I.S.I
    isi_right = np.diff(train_right)
    isi_left = np.diff(train_left)

    if isi_right.size == 0:
        mu_ri = 0
        sigma_ri = 0
    else:
        # mean of I.S.I inside the window
        mu_ri = np.mean(isi_right)
        # var of I.S.I inside the window
        sigma_ri = np.var(isi_right)

    if isi_left.size == 0:
        mu_le = 0
        sigma_le = 0
    else:
        mu_le = np.mean(isi_left)
        sigma_le = np.var(isi_left)

    if (sigma_le > 0) & (sigma_ri > 0):
        s_quad = (sigma_ri / mu_ri**3) * h_sec + (sigma_le / mu_le**3) * h_sec
    else:
        s_quad = 0

    if s_quad == 0:
        difference = 0
    else:
        difference = (count_right - count_left) / np.sqrt(s_quad)

    return difference


def _filter_process(time_step, h, spk, t_final, test_param):
    """
    Given a spike train `spk` and a window size `h`, this function generates
    the `filter derivative process` by evaluating the function `_filter`
    in steps of `time_step`.

    Parameters
    ----------
        h : quantity object
         window's size
        t_final : quantity,
           time on which the window is centered
        spk : list, array or SpikeTrain
           spike train to analyze
        time_step : quantity object, time step at which the windows are slided
          resolution
        test_param : matrix, the means of the first row list of `h`,
                    the second row Empirical and the third row variances of
                    the limit processes `Lh` are used to normalize the number
                    of elements inside the windows

    Returns
    -------
        time_domain : numpy array
                   time domain of the `filter derivative process`
        filter_process : array,
                      values of the `filter derivative process`
    """

    u = 1 * pq.s

    try:
        h_sec = h.rescale(u).magnitude
    except AttributeError:
        raise ValueError("h must be a time quantity")
    try:
        t_final_sec = t_final.rescale(u).magnitude
    except AttributeError:
        raise ValueError("t_final must be a time quanity")
    try:
        dt_sec = time_step.rescale(u).magnitude
    except AttributeError:
        raise ValueError("time_step must be a time quantity")
    # domain of the process
    time_domain = np.arange(h_sec, t_final_sec - h_sec, dt_sec)
    filter_trajectrory = []
    # taken from the function used to generate the threshold
    emp_mean_h = test_param[1][test_param[0] == h]
    emp_var_h = test_param[2][test_param[0] == h]

    for t in time_domain:
        filter_trajectrory.append(_filter(t * u, h, spk))

    filter_trajectrory = np.asanyarray(filter_trajectrory)
    # ordered normalization to give each process the same impact on the max
    filter_process = (
        np.abs(filter_trajectrory) - emp_mean_h) / np.sqrt(emp_var_h)

    return time_domain, filter_process



import hashlib
import tempfile
import warnings
import ssl

from elephant import _get_version
from pathlib import Path
from urllib.request import urlretrieve, urlopen
from urllib.error import HTTPError, URLError
from zipfile import ZipFile
from os import environ, getenv

from tqdm import tqdm

ELEPHANT_TMP_DIR = Path(tempfile.gettempdir()) / "elephant"


class TqdmUpTo(tqdm):
    """
    Provides `update_to(n)` which uses `tqdm.update(delta_n)`.
    Original implementation:
    https://github.com/tqdm/tqdm/blob/master/examples/tqdm_wget.py
    """

    def update_to(self, b=1, bsize=1, tsize=None):
        """
        b : int, optional
            Number of blocks transferred so far [default: 1].
        bsize : int, optional
            Size of each block (in tqdm units) [default: 1].
        tsize : int, optional
            Total size (in tqdm units). If [default: None] remains unchanged.
        """
        if tsize is not None:
            self.total = tsize
        self.update(b * bsize - self.n)  # will also set self.n = b * bsize


def calculate_md5(filepath, chunk_size=1024 * 1024):
    md5 = hashlib.md5()
    with open(filepath, 'rb') as f:
        for chunk in iter(lambda: f.read(chunk_size), b''):
            md5.update(chunk)
    return md5.hexdigest()


def check_integrity(filepath, md5):
    if not Path(filepath).exists() or md5 is None:
        return False
    return calculate_md5(filepath) == md5


def download(url, filepath=None, checksum=None, verbose=True):
    if filepath is None:
        filename = url.split('/')[-1]
        filepath = ELEPHANT_TMP_DIR / filename
    filepath = Path(filepath)
    if check_integrity(filepath, md5=checksum):
        return filepath
    folder = filepath.absolute().parent
    folder.mkdir(exist_ok=True)
    desc = f"Downloading {url} to '{filepath}'"
    with TqdmUpTo(unit='B', unit_scale=True, unit_divisor=1024, miniters=1,
                  desc=desc, disable=not verbose) as t:
        try:
            urlretrieve(url, filename=filepath, reporthook=t.update_to)
        except URLError:
            # do not authenticate SSL certificate
            ssl._create_default_https_context = ssl._create_unverified_context
            urlretrieve(url, filename=filepath, reporthook=t.update_to)
    return filepath


def download_datasets(repo_path, filepath=None, checksum=None,
                      verbose=True):
    r"""
        This function can be used to download files from elephant-data using
        only the path relative to the root of the elephant-data repository.
        The default URL used, points to elephants corresponding release of
        elephant-data.
        Different versions of the elephant package may require different
        versions of elephant-data.
        e.g. the follwoing URLs:
        -  https://web.gin.g-node.org/NeuralEnsemble/elephant-data/raw/0.0.1
           points to release v0.0.1.
        -  https://web.gin.g-node.org/NeuralEnsemble/elephant-data/raw/master
           always points to the latest state of elephant-data.
        -  https://datasets.python-elephant.org/
           points to the root of elephant data

        To change this URL, use the environment variable `ELEPHANT_DATA_URL`.
        When using data, which is not yet contained in the master branch or a
        release of elephant data, e.g. during development, this variable can
        be used to change the default URL.
        For example to use data on branch `multitaper`, change the
        `ELEPHANT_DATA_URL` to
        https://web.gin.g-node.org/NeuralEnsemble/elephant-data/raw/multitaper.
        For a complete example, see Examples section.

        Parameters
        ----------
        repo_path : str
            String denoting the path relative to elephant-data repository root
        filepath : str, optional
            Path to temporary folder where the downloaded files will be stored
        checksum : str, otpional
            Checksum to verify dara integrity after download
        verbose : bool, optional
            Whether to disable the entire progressbar wrapper [].
            If set to None, disable on non-TTY.
            Default: True

        Returns
        -------
        filepath : pathlib.Path
            Path to downloaded files.


        Notes
        -----
        The default URL always points to elephant-data. Please
        do not change its value. For development purposes use the environment
        variable 'ELEPHANT_DATA_URL'.

        Examples
        --------
        The following example downloads a file from elephant-data branch
        'multitaper', by setting the environment variable to the branch URL:

        >>> import os
        >>> from elephant.datasets import download_datasets
        >>> os.environ["ELEPHANT_DATA_URL"] = "https://web.gin.g-node.org/NeuralEnsemble/elephant-data/raw/multitaper" # noqa
        >>> download_datasets("unittest/spectral/multitaper_psd/data/time_series.npy") # doctest: +SKIP
        PosixPath('/tmp/elephant/time_series.npy')
        """

    # this url redirects to the current location of elephant-data
    url_to_root = "https://datasets.python-elephant.org/"

    # get URL to corresponding version of elephant data
    # (version elephant is equal to version elephant-data)
    default_url = url_to_root + f"raw/v{_get_version()}"

    if 'ELEPHANT_DATA_URL' not in environ:  # user did not set URL
        # is 'version-URL' available? (not for elephant development version)
        try:
            urlopen(default_url+'/README.md')

        except HTTPError as error:
            # if corresponding elephant-data version is not found,
            # use latest commit of elephant-data
            default_url = url_to_root + f"raw/master"

            warnings.warn(f"No corresponding version of elephant-data found.\n"
                          f"Elephant version: {_get_version()}. "
                          f"Data URL:{error.url}, error: {error}.\n"
                          f"Using elephant-data latest instead (This is "
                          f"expected for elephant development versions).")

        except URLError as error:
            # if verification of SSL certificate fails, do not verify cert
            try:  # try again without certificate verification
                ctx = ssl._create_unverified_context()
                ctx.check_hostname = True
                urlopen(default_url + '/README.md')
            except HTTPError:  # e.g. 404
                default_url = url_to_root + f"raw/master"

            warnings.warn(f"Data URL:{default_url}, error: {error}."
                          f"{error.reason}")

    url = f"{getenv('ELEPHANT_DATA_URL', default_url)}/{repo_path}"

    return download(url, filepath, checksum, verbose)


def unzip(filepath, outdir=ELEPHANT_TMP_DIR, verbose=True):
    with ZipFile(filepath) as zfile:
        zfile.extractall(path=outdir)
    if verbose:
        print(f"Extracted {filepath} to {outdir}")



# -*- coding: utf-8 -*-
"""
Unit tests for the phase analysis module.

:copyright: Copyright 2014-2023 by the Elephant team, see `doc/authors.rst`.
:license: Modified BSD, see LICENSE.txt for details.
"""
from __future__ import division, print_function

import unittest

import numpy as np
import quantities as pq
import scipy.io
from neo import SpikeTrain, AnalogSignal
from numpy.ma.testutils import assert_allclose

import elephant.phase_analysis
from elephant.datasets import download_datasets


class SpikeTriggeredPhaseTestCase(unittest.TestCase):

    def setUp(self):
        tlen0 = 100 * pq.s
        f0 = 20. * pq.Hz
        fs0 = 1 * pq.ms
        t0 = np.arange(
            0, tlen0.rescale(pq.s).magnitude,
            fs0.rescale(pq.s).magnitude) * pq.s
        self.anasig0 = AnalogSignal(
            np.sin(2 * np.pi * (f0 * t0).simplified.magnitude),
            units=pq.mV, t_start=0 * pq.ms, sampling_period=fs0)
        self.st0 = SpikeTrain(
            np.arange(50, tlen0.rescale(pq.ms).magnitude - 50, 50) * pq.ms,
            t_start=0 * pq.ms, t_stop=tlen0)
        self.st1 = SpikeTrain(
            [100., 100.1, 100.2, 100.3, 100.9, 101.] * pq.ms,
            t_start=0 * pq.ms, t_stop=tlen0)

    def test_perfect_locking_one_spiketrain_one_signal(self):
        phases, amps, times = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            self.st0,
            interpolate=True)

        assert_allclose(phases[0], - np.pi / 2.)
        assert_allclose(amps[0], 1, atol=0.1)
        assert_allclose(times[0].magnitude, self.st0.magnitude)
        self.assertEqual(len(phases[0]), len(self.st0))
        self.assertEqual(len(amps[0]), len(self.st0))
        self.assertEqual(len(times[0]), len(self.st0))

    def test_perfect_locking_many_spiketrains_many_signals(self):
        phases, amps, times = elephant.phase_analysis.spike_triggered_phase(
            [
                elephant.signal_processing.hilbert(self.anasig0),
                elephant.signal_processing.hilbert(self.anasig0)],
            [self.st0, self.st0],
            interpolate=True)

        assert_allclose(phases[0], -np.pi / 2.)
        assert_allclose(amps[0], 1, atol=0.1)
        assert_allclose(times[0].magnitude, self.st0.magnitude)
        self.assertEqual(len(phases[0]), len(self.st0))
        self.assertEqual(len(amps[0]), len(self.st0))
        self.assertEqual(len(times[0]), len(self.st0))

    def test_perfect_locking_one_spiketrains_many_signals(self):
        phases, amps, times = elephant.phase_analysis.spike_triggered_phase(
            [
                elephant.signal_processing.hilbert(self.anasig0),
                elephant.signal_processing.hilbert(self.anasig0)],
            [self.st0],
            interpolate=True)

        assert_allclose(phases[0], -np.pi / 2.)
        assert_allclose(amps[0], 1, atol=0.1)
        assert_allclose(times[0].magnitude, self.st0.magnitude)
        self.assertEqual(len(phases[0]), len(self.st0))
        self.assertEqual(len(amps[0]), len(self.st0))
        self.assertEqual(len(times[0]), len(self.st0))

    def test_perfect_locking_many_spiketrains_one_signal(self):
        phases, amps, times = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            [self.st0, self.st0],
            interpolate=True)

        assert_allclose(phases[0], -np.pi / 2.)
        assert_allclose(amps[0], 1, atol=0.1)
        assert_allclose(times[0].magnitude, self.st0.magnitude)
        self.assertEqual(len(phases[0]), len(self.st0))
        self.assertEqual(len(amps[0]), len(self.st0))
        self.assertEqual(len(times[0]), len(self.st0))

    def test_interpolate(self):
        phases_int, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            self.st1,
            interpolate=True)

        self.assertLess(phases_int[0][0], phases_int[0][1])
        self.assertLess(phases_int[0][1], phases_int[0][2])
        self.assertLess(phases_int[0][2], phases_int[0][3])
        self.assertLess(phases_int[0][3], phases_int[0][4])
        self.assertLess(phases_int[0][4], phases_int[0][5])

        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            self.st1,
            interpolate=False)

        self.assertEqual(phases_noint[0][0], phases_noint[0][1])
        self.assertEqual(phases_noint[0][1], phases_noint[0][2])
        self.assertEqual(phases_noint[0][2], phases_noint[0][3])
        self.assertEqual(phases_noint[0][3], phases_noint[0][4])
        self.assertNotEqual(phases_noint[0][4], phases_noint[0][5])

        # Verify that when using interpolation and the spike sits on the sample
        # of the Hilbert transform, this is the same result as when not using
        # interpolation with a spike slightly to the right
        self.assertEqual(phases_noint[0][2], phases_int[0][0])
        self.assertEqual(phases_noint[0][4], phases_int[0][0])

    def test_inconsistent_numbers_spiketrains_hilbert(self):
        self.assertRaises(
            ValueError, elephant.phase_analysis.spike_triggered_phase,
            [
                elephant.signal_processing.hilbert(self.anasig0),
                elephant.signal_processing.hilbert(self.anasig0)],
            [self.st0, self.st0, self.st0], False)

        self.assertRaises(
            ValueError, elephant.phase_analysis.spike_triggered_phase,
            [
                elephant.signal_processing.hilbert(self.anasig0),
                elephant.signal_processing.hilbert(self.anasig0)],
            [self.st0, self.st0, self.st0], False)

    def test_spike_earlier_than_hilbert(self):
        # This is a spike clearly outside the bounds
        st = SpikeTrain(
            [-50, 50],
            units='s', t_start=-100 * pq.s, t_stop=100 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 1)

        # This is a spike right on the border (start of the signal is at 0s,
        # spike sits at t=0s). By definition of intervals in
        # Elephant (left borders inclusive, right borders exclusive), this
        # spike is to be considered.
        st = SpikeTrain(
            [0, 50],
            units='s', t_start=-100 * pq.s, t_stop=100 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 2)

    def test_spike_later_than_hilbert(self):
        # This is a spike clearly outside the bounds
        st = SpikeTrain(
            [1, 250],
            units='s', t_start=-1 * pq.s, t_stop=300 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 1)

        # This is a spike right on the border (length of the signal is 100s,
        # spike sits at t=100s). However, by definition of intervals in
        # Elephant (left borders inclusive, right borders exclusive), this
        # spike is not to be considered.
        st = SpikeTrain(
            [1, 100],
            units='s', t_start=-1 * pq.s, t_stop=200 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 1)

    # This test handles the correct dealing with input signals that have
    # different time units, including a CompoundUnit
    def test_regression_269(self):
        # This is a spike train on a 30KHz sampling, one spike at 1s, one just
        # before the end of the signal
        cu = pq.CompoundUnit("1/30000.*s")
        st = SpikeTrain(
            [30000., (self.anasig0.t_stop - 1 * pq.s).rescale(cu).magnitude],
            units=pq.CompoundUnit("1/30000.*s"),
            t_start=-1 * pq.s, t_stop=300 * pq.s)
        phases_noint, _, _ = elephant.phase_analysis.spike_triggered_phase(
            elephant.signal_processing.hilbert(self.anasig0),
            st,
            interpolate=False)
        self.assertEqual(len(phases_noint[0]), 2)


class MeanVectorTestCase(unittest.TestCase):
    def setUp(self):
        self.tolerance = 1e-15
        self.n_samples = 200
        # create a sample with all values equal to a random phase-lock phi
        self.lock_value_phi = np.random.uniform(-np.pi, np.pi, 1)
        self.dataset1 = np.ones(self.n_samples) * self.lock_value_phi
        # create a evenly spaced / uniform distribution
        self.dataset2 = np.arange(0, 2 * np.pi, (2 * np.pi) / self.n_samples)
        # create a random distribution
        self.dataset3 = np.random.uniform(-np.pi, np.pi, self.n_samples)

    def testMeanVector_direction_is_phi_and_length_is_1(self):
        """
        Test if the mean vector length is 1 and if the mean direction is phi
        for a sample with all phases equal to phi on the unit circle.

        """
        theta_bar_1, r_1 = elephant.phase_analysis.mean_phase_vector(
            self.dataset1)
        # mean direction must be phi
        self.assertAlmostEqual(theta_bar_1, self.lock_value_phi,
                               delta=self.tolerance)
        # mean vector length must be almost equal 1
        self.assertAlmostEqual(r_1, 1, delta=self.tolerance)

    def testMeanVector_length_is_0(self):
        """
        Test if the mean vector length  is 0 for a evenly spaced distribution
        on the unit circle.
        """
        theta_bar_2, r_2 = elephant.phase_analysis.mean_phase_vector(
            self.dataset2)
        # mean vector length must be almost equal 0
        self.assertAlmostEqual(r_2, 0, delta=self.tolerance)

    def testMeanVector_ranges_of_direction_and_length(self):
        """
        Test if the range of the mean vector direction follows numpy standard
        and is within (-pi, pi].
        Test if the range of the mean vector length is within [0, 1].
        """
        theta_bar_3, r_3 = elephant.phase_analysis.mean_phase_vector(
            self.dataset3)
        # mean vector direction
        self.assertTrue(-np.pi < theta_bar_3 <= np.pi)
        # mean vector length
        self.assertTrue(0 <= r_3 <= 1)


class PhaseDifferenceTestCase(unittest.TestCase):
    def setUp(self):
        self.tolerance = 1e-15
        self.n_samples = 200

    def testPhaseDifference_in_range_minus_pi_to_pi(self):
        """
        Test if the range of the phase difference is within [-pi, pi] for
        random pairs of alpha and beta.
        """
        alpha = np.random.uniform(-np.pi, np.pi, self.n_samples)
        beta = np.random.uniform(-np.pi, np.pi, self.n_samples)

        phase_diff = elephant.phase_analysis.phase_difference(alpha, beta)
        self.assertTrue((-np.pi <= phase_diff).all()
                        and (phase_diff <= np.pi).all())

    def testPhaseDifference_is_delta(self):
        """
        Test if the phase difference is random delta for random pairs of
        alpha and beta, where beta is a copy of alpha shifted by delta.
        """
        delta = np.random.uniform(-np.pi, np.pi, self.n_samples)
        alpha = np.random.uniform(-np.pi, np.pi, self.n_samples)
        _beta = alpha - delta
        beta = np.arctan2(np.sin(_beta), np.cos(_beta))

        phase_diff = elephant.phase_analysis.phase_difference(alpha, beta)
        np.testing.assert_allclose(phase_diff, delta, atol=1e-10)


class PhaseLockingValueTestCase(unittest.TestCase):
    def setUp(self):
        self.tolerance = 1e-15
        self.phase_shift = np.pi / 4
        self.num_time_points = 1000
        self.num_trials = 100

        # create two random uniform distributions (all trials are identical)
        self.signal_x = \
            np.full([self.num_trials, self.num_time_points],
                    np.random.uniform(-np.pi, np.pi, self.num_time_points))
        self.signal_y = \
            np.full([self.num_trials, self.num_time_points],
                    np.random.uniform(-np.pi, np.pi, self.num_time_points))

        # create two random uniform distributions, where all trails are random
        self.random_x = np.random.uniform(
            -np.pi, np.pi, (1000, self.num_time_points))
        self.random_y = np.random.uniform(
            -np.pi, np.pi, (1000, self.num_time_points))

        # simple samples of different shapes to assert ErrorRaising
        self.simple_x = np.array([[0, -np.pi, np.pi], [0, -np.pi, np.pi]])
        self.simple_y = np.array([0, -np.pi, np.pi])
        self.simple_z = np.array([0, np.pi, np.pi / 2, -np.pi])

    def testPhaseLockingValue_identical_signals_both_identical_trials(self):
        """
        Test if the PLV's are 1, when 2 identical signals with identical
        trials are passed. PLV's needed to be 1, due to the constant phase
        difference of 0 across trials at each time-point.
        """
        list1_plv_t = \
            elephant.phase_analysis.phase_locking_value(self.signal_x,
                                                        self.signal_x)
        target_plv_r_is_one = np.ones_like(list1_plv_t)
        np.testing.assert_allclose(list1_plv_t, target_plv_r_is_one,
                                   self.tolerance)

    def testPhaseLockingValue_different_signals_both_identical_trials(self):
        """
        Test if the PLV's are 1, when 2 different signals are passed, where
        within each signal the trials are identical. PLV's needed to be 1,
        due to a constant phase difference across trials, which may vary for
        different time-points.
        """
        list2_plv_t = elephant.phase_analysis.phase_locking_value(
            self.signal_x, self.signal_y)
        target_plv_r_is_one = np.ones_like(list2_plv_t)
        np.testing.assert_allclose(list2_plv_t, target_plv_r_is_one,
                                   atol=3e-15)

    def testPhaseLockingValue_different_signals_both_different_trials(self):
        """
        Test if the PLV's are close to 0, when 2 different signals are passed,
        where both have different trials, which are all randomly distributed.
        The PLV's needed to be close to 0, do to a random
        phase difference across trials for each time-point.
        """
        list3_plv_t = elephant.phase_analysis.phase_locking_value(
            self.random_x, self.random_y)
        target_plv_is_zero = np.zeros_like(list3_plv_t)
        # use default value from np.allclose() for atol=1e-8 to prevent failure
        np.testing.assert_allclose(list3_plv_t, target_plv_is_zero,
                                   rtol=1e-2, atol=1.1e-1)

    def testPhaseLockingValue_raise_Error_if_trial_number_is_different(self):
        """
        Test if a ValueError is raised, when the signals have different
        number of trails.
        """
        # different numbers of trails
        np.testing.assert_raises(
            ValueError, elephant.phase_analysis.phase_locking_value,
            self.simple_x, self.simple_y)

    def testPhaseLockingValue_raise_Error_if_trial_lengths_are_different(self):
        """
        Test if a ValueError is raised, when within a trail-pair of the signals
        the trial-lengths are different.
        """
        # different lengths in a trail pair
        np.testing.assert_raises(
            ValueError, elephant.phase_analysis.phase_locking_value,
            self.simple_y, self.simple_z)


class WeightedPhaseLagIndexTestCase(unittest.TestCase):
    files_to_download_ground_truth = None
    files_to_download_artificial = None
    files_to_download_real = None

    @classmethod
    def setUpClass(cls):
        np.random.seed(73)

        # The files from G-Node GIN 'elephant-data' repository will be
        # downloaded once into a local temporary directory
        # and then loaded/ read for each test function individually.

        # REAL DATA
        real_data_path = "unittest/phase_analysis/weighted_phase_lag_index/" \
                         "data/wpli_real_data"
        cls.files_to_download_real = (
            ("i140703-001_ch01_slice_TS_ON_to_GO_ON_correct_trials.mat",
             "0e76454c58208cab710e672d04de5168"),
            ("i140703-001_ch02_slice_TS_ON_to_GO_ON_correct_trials.mat",
             "b06059e5222e91eb640caad0aba15b7f"),
            ("i140703-001_cross_spectrum_of_channel_1_and_2_of_slice_"
             "TS_ON_to_GO_ON_corect_trials.mat",
             "2687ef63a4a456971a5dcc621b02e9a9")
        )
        for filename, checksum in cls.files_to_download_real:
            # files will be downloaded to ELEPHANT_TMP_DIR
            cls.tmp_path = download_datasets(
                f"{real_data_path}/{filename}", checksum=checksum)
        # ARTIFICIAL DATA
        artificial_data_path = "unittest/phase_analysis/" \
            "weighted_phase_lag_index/data/wpli_specific_artificial_dataset"
        cls.files_to_download_artificial = (
            ("artificial_LFPs_1.mat", "4b99b15f89c0b9a0eb6fc14e9009436f"),
            ("artificial_LFPs_2.mat", "7144976b5f871fa62f4a831f530deee4"),
        )
        for filename, checksum in cls.files_to_download_artificial:
            # files will be downloaded to ELEPHANT_TMP_DIR
            cls.tmp_path = download_datasets(
                f"{artificial_data_path}/{filename}", checksum=checksum)
        # GROUND TRUTH DATA
        ground_truth_data_path = "unittest/phase_analysis/" \
                        "weighted_phase_lag_index/data/wpli_ground_truth"
        cls.files_to_download_ground_truth = (
            ("ground_truth_WPLI_from_ft_connectivity_wpli_"
             "with_real_LFPs_R2G.csv", "4d9a7b7afab7d107023956077ab11fef"),
            ("ground_truth_WPLI_from_ft_connectivity_wpli_"
             "with_artificial_LFPs.csv", "92988f475333d7badbe06b3f23abe494"),
        )
        for filename, checksum in cls.files_to_download_ground_truth:
            # files will be downloaded into ELEPHANT_TMP_DIR
            cls.tmp_path = download_datasets(
                f"{ground_truth_data_path}/{filename}", checksum=checksum)

    def setUp(self):
        self.tolerance = 1e-15

        # load real/artificial LFP-dataset for ground-truth consistency checks
        # real LFP-dataset
        dataset1_real = scipy.io.loadmat(
            f"{self.tmp_path.parent}/{self.files_to_download_real[0][0]}",
            squeeze_me=True)
        dataset2_real = scipy.io.loadmat(
            f"{self.tmp_path.parent}/{self.files_to_download_real[1][0]}",
            squeeze_me=True)

        # get relevant values
        self.lfps1_real = dataset1_real['lfp_matrix'] * pq.uV
        self.sf1_real = dataset1_real['sf'] * pq.Hz
        self.lfps2_real = dataset2_real['lfp_matrix'] * pq.uV
        self.sf2_real = dataset2_real['sf'] * pq.Hz
        # create AnalogSignals from the real dataset
        self.lfps1_real_AnalogSignal = AnalogSignal(
            signal=self.lfps1_real, sampling_rate=self.sf1_real)
        self.lfps2_real_AnalogSignal = AnalogSignal(
            signal=self.lfps2_real, sampling_rate=self.sf2_real)

        # artificial LFP-dataset
        dataset1_artificial = scipy.io.loadmat(
            f"{self.tmp_path.parent}/"
            f"{self.files_to_download_artificial[0][0]}", squeeze_me=True)
        dataset2_artificial = scipy.io.loadmat(
            f"{self.tmp_path.parent}/"
            f"{self.files_to_download_artificial[1][0]}", squeeze_me=True)
        # get relevant values
        self.lfps1_artificial = dataset1_artificial['lfp_matrix'] * pq.uV
        self.sf1_artificial = dataset1_artificial['sf'] * pq.Hz
        self.lfps2_artificial = dataset2_artificial['lfp_matrix'] * pq.uV
        self.sf2_artificial = dataset2_artificial['sf'] * pq.Hz
        # create AnalogSignals from the artificial dataset
        self.lfps1_artificial_AnalogSignal = AnalogSignal(
            signal=self.lfps1_artificial, sampling_rate=self.sf1_artificial)
        self.lfps2_artificial_AnalogSignal = AnalogSignal(
            signal=self.lfps2_artificial, sampling_rate=self.sf2_artificial)

        # load ground-truth reference calculated by:
        # Matlab package 'FieldTrip': ft_connectivity_wpli()
        self.wpli_ground_truth_ft_connectivity_wpli_real = np.loadtxt(
            f"{self.tmp_path.parent}/"
            f"{self.files_to_download_ground_truth[0][0]}",
            delimiter=',', dtype=np.float64)
        self.wpli_ground_truth_ft_connectivity_artificial = np.loadtxt(
            f"{self.tmp_path.parent}/"
            f"{self.files_to_download_ground_truth[1][0]}",
            delimiter=',', dtype=np.float64)

    def test_WPLI_ground_truth_consistency_real_LFP_dataset(self):
        """
        Test if the WPLI is consistent with the reference implementation
        ft_connectivity_wpli() of the MATLAB-package FieldTrip using
        LFP-dataset cuttings from the multielectrode-grasp  G-Node GIN
        repository, which can be found here:
        https://doi.gin.g-node.org/10.12751/g-node.f83565/
        The cutting was performed with this python-script:
        multielectrode_grasp_i140703-001_cutting_script_TS_ON_to_GO_ON.py
        which is available on https://gin.g-node.org/INM-6/elephant-data
        in folder validation/phase_analysis/weighted_phase_lag_index/scripts,
        where also the MATLAB-script for ground-truth generation is located.
        """
        # Quantity-input
        with self.subTest(msg="Quantity input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_real, self.lfps2_real, self.sf1_real)
            np.testing.assert_allclose(
                wpli, self.wpli_ground_truth_ft_connectivity_wpli_real,
                equal_nan=True)
        # np.array-input
        with self.subTest(msg="np.array input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_real.magnitude, self.lfps2_real.magnitude,
                self.sf1_real)
            np.testing.assert_allclose(
                wpli, self.wpli_ground_truth_ft_connectivity_wpli_real,
                equal_nan=True)
        # neo.AnalogSignal-input
        with self.subTest(msg="neo.AnalogSignal input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_real_AnalogSignal, self.lfps2_real_AnalogSignal)
            np.testing.assert_allclose(
                wpli, self.wpli_ground_truth_ft_connectivity_wpli_real,
                equal_nan=True)

    def test_WPLI_ground_truth_consistency_artificial_LFP_dataset(self):
        """
        Test if the WPLI is consistent with the ground truth generated with
        multi-sine artificial LFP-datasets.
        The generation was performed with this python-script:
        generate_artificial_datasets_for_ground_truth_of_wpli.py
        which is available on https://gin.g-node.org/INM-6/elephant-data
        in folder validation/phase_analysis/weighted_phase_lag_index/scripts,
        where also the MATLAB-script for ground-truth generation is located.
        """
        # Quantity-input
        with self.subTest(msg="Quantity input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial, self.lfps2_artificial,
                self.sf1_artificial, absolute_value=False)
            np.testing.assert_allclose(
                wpli, self.wpli_ground_truth_ft_connectivity_artificial,
                atol=1e-14, rtol=1e-12, equal_nan=True)
        # np.array-input
        with self.subTest(msg="np.array input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial.magnitude,
                self.lfps2_artificial.magnitude, self.sf1_artificial,
                absolute_value=False)
            np.testing.assert_allclose(
                wpli, self.wpli_ground_truth_ft_connectivity_artificial,
                atol=1e-14, rtol=1e-12, equal_nan=True)
        # neo.AnalogSignal-input
        with self.subTest(msg="neo.AnalogSignal input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial_AnalogSignal,
                self.lfps2_artificial_AnalogSignal, absolute_value=False)
            np.testing.assert_allclose(
                wpli, self.wpli_ground_truth_ft_connectivity_artificial,
                atol=1e-14, rtol=1e-12, equal_nan=True)

    def test_WPLI_is_zero(self):
        """
        Test if WPLI is close to zero at frequency f=70Hz for the multi-sine
        artificial LFP dataset. White noise prevents arbitrary approximation.
        """
        # Quantity-input
        with self.subTest(msg="Quantity input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial, self.lfps2_artificial,
                self.sf1_artificial, absolute_value=False)
            np.testing.assert_allclose(
                wpli[freq == 70], 0, atol=0.004, rtol=self.tolerance)
        # np.array-input
        with self.subTest(msg="np.array input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial.magnitude,
                self.lfps2_artificial.magnitude, self.sf1_artificial,
                absolute_value=False)
            np.testing.assert_allclose(
                wpli[freq == 70], 0, atol=0.004, rtol=self.tolerance)
        # neo.AnalogSignal-input
        with self.subTest(msg="neo.AnalogSignal input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial_AnalogSignal,
                self.lfps2_artificial_AnalogSignal, absolute_value=False)
            np.testing.assert_allclose(
                wpli[freq == 70], 0, atol=0.004, rtol=self.tolerance)

    def test_WPLI_is_one(self):
        """
        Test if WPLI is one at frequency f=16Hz and 36Hz for the multi-sine
        artificial LFP dataset.
        """
        # Quantity-input
        with self.subTest(msg="Quantity input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial, self.lfps2_artificial,
                self.sf1_artificial, absolute_value=False)
            mask = ((freq == 16) | (freq == 36))
            np.testing.assert_allclose(
                wpli[mask], 1, atol=self.tolerance, rtol=self.tolerance)
        # np.array-input
        with self.subTest(msg="np.array input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial.magnitude,
                self.lfps2_artificial.magnitude, self.sf1_artificial,
                absolute_value=False)
            mask = ((freq == 16) | (freq == 36))
            np.testing.assert_allclose(
                wpli[mask], 1, atol=self.tolerance, rtol=self.tolerance)
        # neo.AnalogSignal-input
        with self.subTest(msg="neo.AnalogSignal input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial_AnalogSignal,
                self.lfps2_artificial_AnalogSignal, absolute_value=False)
            mask = ((freq == 16) | (freq == 36))
            np.testing.assert_allclose(
                wpli[mask], 1, atol=self.tolerance, rtol=self.tolerance)

    def test_WPLI_is_minus_one(self):
        """
        Test if WPLI is minus one at frequency f=52Hz and 100Hz
        for the multi-sine artificial LFP dataset.
        """
        # Quantity-input
        with self.subTest(msg="Quantity input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial, self.lfps2_artificial,
                self.sf1_artificial, absolute_value=False)
            mask = ((freq == 52) | (freq == 100))
            np.testing.assert_allclose(
                wpli[mask], -1, atol=self.tolerance, rtol=self.tolerance)
        # np.array-input
        with self.subTest(msg="np.array input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial.magnitude,
                self.lfps2_artificial.magnitude, self.sf1_artificial,
                absolute_value=False)
            np.testing.assert_allclose(
                wpli[mask], -1, atol=self.tolerance, rtol=self.tolerance)
        # neo.AnalogSignal-input
        with self.subTest(msg="neo.AnalogSignal input"):
            freq, wpli = elephant.phase_analysis.weighted_phase_lag_index(
                self.lfps1_artificial_AnalogSignal,
                self.lfps2_artificial_AnalogSignal, absolute_value=False)
            np.testing.assert_allclose(
                wpli[mask], -1, atol=self.tolerance, rtol=self.tolerance)

    def test_WPLI_raises_error_if_signals_have_different_shapes(self):
        """
        Test if WPLI raises a ValueError, when the signals have different
        number of trails or different trial lengths.
        """
        # simple samples of different shapes to assert ErrorRaising
        trials2_length3 = np.array([[0, -1, 1], [0, -1, 1]]) * pq.uV
        trials1_length3 = np.array([[0, -1, 1]]) * pq.uV
        trials1_length4 = np.array([[0, 1, 1 / 2, -1]]) * pq.uV
        sampling_frequency = 250 * pq.Hz
        trials2_length3_analogsignal = AnalogSignal(
            signal=trials2_length3, sampling_rate=sampling_frequency)
        trials1_length3_analogsignal = AnalogSignal(
            signal=trials1_length3, sampling_rate=sampling_frequency)
        trials1_length4_analogsignal = AnalogSignal(
            signal=trials1_length4, sampling_rate=sampling_frequency)

        # different numbers of trails
        with self.subTest(msg="diff. trial numbers & Quantity input"):
            np.testing.assert_raises(
                ValueError, elephant.phase_analysis.weighted_phase_lag_index,
                trials2_length3, trials1_length3, sampling_frequency)
        with self.subTest(msg="diff. trial numbers & np.array input"):
            np.testing.assert_raises(
                ValueError, elephant.phase_analysis.weighted_phase_lag_index,
                trials2_length3.magnitude, trials1_length3.magnitude,
                sampling_frequency)
        with self.subTest(msg="diff. trial numbers & neo.AnalogSignal input"):
            np.testing.assert_raises(
                ValueError, elephant.phase_analysis.weighted_phase_lag_index,
                trials2_length3_analogsignal, trials1_length3_analogsignal)
        # different lengths in a trail pair
        with self.subTest(msg="diff. trial lengths & Quantity input"):
            np.testing.assert_raises(
                ValueError, elephant.phase_analysis.weighted_phase_lag_index,
                trials1_length3, trials1_length4, sampling_frequency)
        with self.subTest(msg="diff. trial lengths & np.array input"):
            np.testing.assert_raises(
                ValueError, elephant.phase_analysis.weighted_phase_lag_index,
                trials1_length3.magnitude, trials1_length4.magnitude,
                sampling_frequency)
        with self.subTest(msg="diff. trial lengths & neo.AnalogSignal input"):
            np.testing.assert_raises(
                ValueError, elephant.phase_analysis.weighted_phase_lag_index,
                trials1_length3_analogsignal, trials1_length4_analogsignal)

    @staticmethod
    def test_WPLI_raises_error_if_AnalogSignals_have_diff_sampling_rate():
        """
        Test if WPLI raises a ValueError, when the AnalogSignals have different
        sampling rates.
        """
        signal_x_250_hz = AnalogSignal(signal=np.random.random([40, 2100]),
                                       units=pq.mV, sampling_rate=0.25*pq.kHz)
        signal_y_1000_hz = AnalogSignal(signal=np.random.random([40, 2100]),
                                        units=pq.mV, sampling_rate=1000*pq.Hz)
        np.testing.assert_raises(
            ValueError, elephant.phase_analysis.weighted_phase_lag_index,
            signal_x_250_hz, signal_y_1000_hz)

    def test_WPLI_raises_error_if_sampling_rate_not_given(self):
        """
        Test if WPLI raises a ValueError, when the sampling rate is not given
        for np.array() or Quanitity input.
        """
        signal_x = np.random.random([40, 2100]) * pq.mV
        signal_y = np.random.random([40, 2100]) * pq.mV
        with self.subTest(msg="Quantity-input"):
            np.testing.assert_raises(
                ValueError, elephant.phase_analysis.weighted_phase_lag_index,
                signal_x, signal_y)
        with self.subTest(msg="np.array-input"):
            np.testing.assert_raises(
                ValueError, elephant.phase_analysis.weighted_phase_lag_index,
                signal_x.magnitude, signal_y.magnitude)


if __name__ == '__main__':
    unittest.main()
